﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;

namespace LevelEditor
{
    class Level
    {
        public Level()
        {
            Revert();
        }

        #region Level Attributes

        public class LevelAttributes
        {
            public String levelName = "";
            public String fontFile  = "";
            public int fontSize     = 0; 
            public Color bgColor    = new Color();
            public Rectangle bounds = new Rectangle();
        }

        public class PaddleAttributes
        {
            public String fileName     = "";
            public String shapeFile    = "";
            public Point startPosition = new Point();
            public int width           = 0;
            public int height          = 0;
        }

        public class BallAttributes
        {
            public String fileName     = "";
            public int radius          = 0;
            public Point startPosition = new Point();
            public Point initVelocity  = new Point();
        }

        public class BrickGroupAttributes
        {
            public enum DistrobutionType
		    {
			    DISTROTYPE_RANDOM = 0,
			    DISTROTYPE_POINTS,
			    DISTROTYPE_UNIFORM,
			    DISTROTYPE_UNKNOWN
		    }

            public String name               = "";
            public String Name
            { 
                set { name = value; } 
                get { return name; } 
            }

            public List<String> imageFiles   = new List<String>();
		    public int brickWidth            = 1;
		    public int brickHeight           = 1;
            public int brickHP               = 1;
            public int pointsPerHit          = 0;
            public int pointsPerKill         = 0;

            public DistrobutionType distType = DistrobutionType.DISTROTYPE_UNKNOWN;
    		
		    // valid when distType == RANDOM
		    public Rectangle bounds          = new Rectangle();
		    public uint count                = 0;
		    public bool allowOverlap         = false;

		    // valid when distType == POINTS
            public BindingList<Point> points = new BindingList<Point>();
    		
		    // valid when distType == UNIFORM
		    public Point center              = new Point();
            public int rows                  = 0;
            public int cols                  = 0;
		    public int rowSpace              = 0;
            public int colSpace              = 0;
        }

        public LevelAttributes levelAttribs = new LevelAttributes();
        public PaddleAttributes paddleAttribs = new PaddleAttributes();
        public BallAttributes ballAttribs = new BallAttributes();
        public List<BrickGroupAttributes> brickGroups = new List<BrickGroupAttributes>();

        #endregion

        #region Game Spacific Stuff

        public const int Width = 800;
        public const int Height = 600;
        public const int HalfWidth = Width / 2;
        public const int HalfHeight = Height / 2;

        public static void NormToScreenCoords(float xin, float yin, out int x, out int y)
        {
            x = (int)(xin * HalfWidth);
            y = (int)(yin * HalfHeight);
        }
        public static void ScreenToNormCoords(int xin, int yin, out float x, out float y)
        {
            x = (2.0f * xin) / Width;
            y = (2.0f * yin) / Height;
        }

        // transforms a point defined in the games world space to a point in a windows forms screen space.
        public static void TransformToScreenSpace(Point p, ref Point result)
        {
            result.X = TransformToScreenSpaceX(p.X);
            result.Y = TransformToScreenSpaceY(p.Y);
        }
        public static void TransformToScreenSpaceX(int x, ref int result)
        {
            result = TransformToScreenSpaceX(x);
        }
        public static void TransformToScreenSpaceY(int y, ref int result)
        {
            result = TransformToScreenSpaceY(y);
        }
        public static int TransformToScreenSpaceX(int x)
        {
            return x + HalfWidth;
        }
        public static int TransformToScreenSpaceY(int y)
        {
            return -y + HalfHeight;
        }

        // transforms a point defined in a windows forms screen space to a point in the games world space.
        public static void TransformToGameSpace(Point p, ref Point result)
        {
            result.X = TransformToGameSpaceX(p.X);
            result.Y = TransformToGameSpaceY(p.Y);
        }
        public static void TransformToGameSpaceX(int x, ref int result)
        {
            result = TransformToGameSpaceX(x);
        }
        public static void TransformToGameSpaceY(int y, ref int result)
        {
            result = TransformToGameSpaceY(y);
        }
        public static int TransformToGameSpaceX(int x)
        {
            return x - HalfWidth;
        }
        public static int TransformToGameSpaceY(int y)
        {
            return HalfHeight - y;
        }

        #endregion

        #region Initialization

        public void Revert()
        {
            //
            // levelAttribs
            //
            levelAttribs.bgColor = Color.Black;
            levelAttribs.bounds.X = -Width / 2;
            levelAttribs.bounds.Y = Height / 2;
            levelAttribs.bounds.Width = Width;
            levelAttribs.bounds.Height = Height;
            levelAttribs.fontFile = "Arial.ttf";
            levelAttribs.fontSize = 14;
            levelAttribs.levelName = "Default Level Name";

            //
            // paddleAttribs
            //
            paddleAttribs.fileName = "";
            paddleAttribs.height = 0;
            paddleAttribs.width = 0;
            paddleAttribs.startPosition.X = 0;
            paddleAttribs.startPosition.Y = 0;

            //
            // ballAttribs
            //
            ballAttribs.fileName = "";
            ballAttribs.initVelocity.X = 0;
            ballAttribs.initVelocity.Y = 0;
            ballAttribs.radius = 0;
            ballAttribs.startPosition.X = 0;
            ballAttribs.startPosition.Y = 0;

            brickGroups.Clear();

            //
            // Draw data
            //
            levelFont = null;
            ballImage = null;
            paddleImage = null;
            selectedCollection = null;
            selectedBrick = null;
            brickCollections.Clear();
        }

        // Call after level data has been changed.
        public bool Init()
        {
            levelFont = new Font(levelAttribs.fontFile, levelAttribs.fontSize);

            try
            {
                if (ballAttribs.fileName != null && ballAttribs.fileName != "")
                    ballImage = Image.FromFile(System.IO.Path.Combine(DataPaths.Paths["Textures"], ballAttribs.fileName));
                if (paddleAttribs.fileName != null && paddleAttribs.fileName != "")
                    paddleImage = Image.FromFile(System.IO.Path.Combine(DataPaths.Paths["Textures"], paddleAttribs.fileName));
            }
            catch (System.IO.FileNotFoundException e)
            {
                MainWindow.ConsoleWriteLine("ERROR: The image file '" + e.Message + "' could not be found.");
                return false;
            }

            TransformToScreenSpace(paddleAttribs.startPosition, ref paddlePosition);
            paddlePosition.X -= (paddleAttribs.width / 2);
            paddlePosition.Y -= (paddleAttribs.height / 2);

            foreach (BrickGroupAttributes a in brickGroups)
            {
                BrickCollection bc = new BrickCollection(a);
                brickCollections.Add(bc);
            }

            return true;
        }

        #endregion

        #region Status

        private bool changed = false;
        public bool Changed
        {
            get { return changed; }
            set { changed = value; }
        }

        #endregion

        #region Drawing Stuff

        // Data used for drawing the level
        public Font levelFont = null;
        public Image ballImage = null;
        public Image paddleImage = null;
        public Point paddlePosition = new Point();
        public BindingList<BrickCollection> brickCollections = new BindingList<BrickCollection>();
        
        public class Brick
        {
            public Point position = new Point(); // the position in BrickOut screen space
            public Point drawPosition = new Point(); // The position suitable for drawing

            public BrickCollection owner = null;

            public int hp = 0;

            public int attribIndex = -1;

            public String DisplayPoint
            {
                get
                {
                    return "(" + Convert.ToString(position.X) + ", " + Convert.ToString(position.Y) + ")";
                }
            }

            public void MoveTo(Point p)
            {
                owner.SetBrickPosition(this, p);
            }

            public void MoveToGamePosition(Point p)
            {
                owner.SetGameBrickPosition(this, p);
            }
        }

        public class BrickCollection
        {
            public BrickCollection(BrickGroupAttributes att)
            {
                attribs = att;

                foreach (String s in attribs.imageFiles)
                {
                    Image im = Image.FromFile(System.IO.Path.Combine(DataPaths.Paths["Textures"], s));
                    images.Add(im);
                }

                if (attribs.distType == BrickGroupAttributes.DistrobutionType.DISTROTYPE_POINTS)
                {
                    for(int i = 0; i < attribs.points.Count; i++)
                    {
                        Brick b = CreateBrick(attribs.points[i]);
                        b.attribIndex = i;
                        bricks.Add(b);
                    }
                }

                recomputeBounds = true;
            }

            // Attribute properties
            public String Name
            {
                get { return attribs.name; }
                set { attribs.name = value; }
            }

            public int BrickWidth
            {
                get { return attribs.brickWidth; }
                set 
                { 
                    Attribs.brickWidth = value;
                    foreach (Brick b in bricks)
                    {
                        b.drawPosition.X = TransformToScreenSpaceX(b.position.X) - (value / 2);
                    }
                }
            }

            public int BrickHeight
            {
                get { return attribs.brickHeight; }
                set 
                { 
                    Attribs.brickHeight = value;
                    foreach (Brick b in bricks)
                    {
                        b.drawPosition.Y = TransformToScreenSpaceY(b.position.Y) - (value / 2);
                    }
                }
            }

            private BrickGroupAttributes attribs; // referance into the brickGroups list that these bricks were generated from
            public BrickGroupAttributes Attribs
            {
                get { recomputeBounds = true; return attribs; }
                private set { recomputeBounds = true; attribs = value; }
            }

            public List<Image> images = new List<Image>();
            public BindingList<Brick> bricks = new BindingList<Brick>();

            public void SetImage(int index, String name)
            {
                if (index < 0 || index > attribs.imageFiles.Count)
                    throw new ArgumentOutOfRangeException();

                if (index >= images.Count)
                {
                    int amt = index - images.Count + 1;
                    for (int i = 0; i < amt; i++)
                    {
                        String path = System.IO.Path.Combine(DataPaths.Paths["Textures"], attribs.imageFiles[i]);
                        if (System.IO.File.Exists(path))
                            images.Add(Image.FromFile(path));
                        else
                            images.Add(null);
                    }
                }

                if (name != attribs.imageFiles[index])
                {
                    attribs.imageFiles[index] = name;

                    String path = System.IO.Path.Combine(DataPaths.Paths["Textures"], name);
                    if (System.IO.File.Exists(path))
                        images[index] = Image.FromFile(path);
                    else
                        images[index] = null;
                }
            }

            private bool recomputeBounds = true;
            private Rectangle bounds = new Rectangle();
            public Rectangle Bounds 
            {
                get
                {
                    if (recomputeBounds)
                    {
                        ComputeBounds();
                        recomputeBounds = false;
                    }
                    return bounds;
                }

                private set
                {
                    bounds = value;
                }
            }

            public void ComputeBounds()
            {
                if (bricks.Count == 0)
                {
                    bounds.X = 0;
                    bounds.Y = 0;
                    bounds.Width = 0;
                    bounds.Height = 0;
                    return;
                }

                int minx = int.MaxValue;
                int miny = int.MaxValue;
                int maxx = int.MinValue;
                int maxy = int.MinValue;

                foreach (Brick b in bricks)
                {
                    minx = Math.Min(minx, b.drawPosition.X);
                    miny = Math.Min(miny, b.drawPosition.Y);
                    maxx = Math.Max(maxx, b.drawPosition.X);
                    maxy = Math.Max(maxy, b.drawPosition.Y);
                }

                bounds.X = minx;
                bounds.Y = miny;
                bounds.Width = maxx - minx + attribs.brickWidth;
                bounds.Height = maxy - miny + attribs.brickHeight;
            }

            public Brick AddBrickScreenPosition(Point p)
            {
                Point pos = p;
                TransformToGameSpace(p, ref pos);

                return AddBrick(pos);
            }

            public Brick AddBrick(Point p)
            {
                Brick b = CreateBrick(p);
                b.attribIndex = attribs.points.Count;
                attribs.points.Add(p);
                bricks.Add(b);
                return b;
            }

            public void RemoveBrick(Brick b)
            {
                int index = b.attribIndex;
                attribs.points.RemoveAt(index);

                foreach (Brick br in bricks)
                {
                    if (br.attribIndex > index)
                        br.attribIndex--;
                }

                bricks.Remove(b);

                recomputeBounds = true;
            }

            public Brick Pick(Point p)
            {
                foreach (Brick b in bricks)
                {
                    if (p.X > b.drawPosition.X && p.X < (b.drawPosition.X + attribs.brickWidth) &&
                        p.Y > b.drawPosition.Y && p.Y < (b.drawPosition.Y + attribs.brickHeight))
                    {
                        return b;
                    }
                }
                
                return null;
            }

            public void SetBrickPosition(Brick b, Point p)
            {
                b.drawPosition.X = p.X - (attribs.brickWidth / 2);
                b.drawPosition.Y = p.Y - (attribs.brickHeight / 2);

                TransformToGameSpace(p, ref b.position);

                if (b.attribIndex >= 0)
                {
                    attribs.points.RemoveAt(b.attribIndex);
                    attribs.points.Insert(b.attribIndex, b.position);
                }

                recomputeBounds = true;
            }

            public void SetGameBrickPosition(Brick b, Point p)
            {
                b.position = p;

                Point screen = p;
                TransformToScreenSpace(p, ref screen);

                b.drawPosition.X = screen.X - (attribs.brickWidth / 2);
                b.drawPosition.Y = screen.Y - (attribs.brickHeight / 2);

                if (b.attribIndex >= 0)
                {
                    attribs.points.RemoveAt(b.attribIndex);
                    attribs.points.Insert(b.attribIndex, b.position);
                }

                recomputeBounds = true;
            }

            public Brick CreateBrick(Point position)
            {
                Brick b = new Brick();
                b.position = position;
                b.owner = this;

                TransformToScreenSpace(position, ref b.drawPosition);
                b.drawPosition.X -= (attribs.brickWidth / 2);
                b.drawPosition.Y -= (attribs.brickHeight / 2);
                b.hp = attribs.brickHP;

                return b;
            }

            public void Draw(Graphics gfx)
            {
                foreach (Brick b in bricks)
                {
                    if (b.hp > 0)
                    {
                        int index = System.Math.Min(attribs.brickHP - b.hp, images.Count - 1);
                        if (images[index] != null)
                        {
                            gfx.DrawImage(images[index],
                                b.drawPosition.X,
                                b.drawPosition.Y,
                                attribs.brickWidth,
                                attribs.brickHeight);
                        }
                    }
                }
            }
        }

        public Brick Pick(Point p)
        {
            foreach (BrickCollection c in brickCollections)
            {
                Brick b = c.Pick(p);
                if (b != null)
                    return b;
            }

            return null;
        }

        public BrickCollection CreateNewBrickGroup()
        {
            BrickGroupAttributes att = new BrickGroupAttributes();
            att.distType = BrickGroupAttributes.DistrobutionType.DISTROTYPE_POINTS;
            att.name = "Untitled";

            BrickCollection c = new BrickCollection(att);

            brickGroups.Add(att);
            brickCollections.Add(c);

            return c;
        }

        // Brick highlighting
        public void SelectCollection(BrickCollection c)
        {
            selectedCollection = c;
        }

        public void SelectBrick(Brick b)
        {
            selectedBrick = b;
        }

        public Grid grid = null;

        public BrickCollection selectedCollection = null;
        public Brick selectedBrick = null;

        private Pen selectPen = Pens.Green;
        private Brush selectBrush = new HatchBrush(HatchStyle.DottedGrid, Color.FromArgb(90, Color.LightGreen), Color.Transparent);

        private Pen brickSelectPen = Pens.Red;
        private Brush brickSelectBrush = new SolidBrush(Color.FromArgb(90, Color.Red));

        public void Draw(Graphics gfx)
        {
            gfx.Clear(levelAttribs.bgColor);

            if (grid != null)
            {
                grid.Draw(gfx);
            }

            if (paddleImage != null)
            {
                gfx.DrawImage(paddleImage,
                    paddlePosition.X,
                    paddlePosition.Y,
                    paddleAttribs.width,
                    paddleAttribs.height);
            }

            if (ballImage != null)
            {
                gfx.DrawImage(ballImage,
                    paddlePosition.X + (paddleAttribs.width / 2) - ballAttribs.radius,
                    paddlePosition.Y - ballAttribs.radius * 2,
                    ballAttribs.radius * 2,
                    ballAttribs.radius * 2);
            }

            foreach (BrickCollection b in brickCollections)
            {
                b.Draw(gfx);
            }

            if (selectedCollection != null)
            {
                //gfx.FillRectangle(selectBrush, selectedCollection.Bounds);
                gfx.DrawRectangle(selectPen, selectedCollection.Bounds);
            }

            if (selectedBrick != null)
            {
                gfx.FillRectangle(brickSelectBrush, 
                    selectedBrick.drawPosition.X, 
                    selectedBrick.drawPosition.Y,
                    selectedBrick.owner.BrickWidth,
                    selectedBrick.owner.BrickHeight);
                gfx.DrawRectangle(brickSelectPen,
                    selectedBrick.drawPosition.X,
                    selectedBrick.drawPosition.Y,
                    selectedBrick.owner.BrickWidth,
                    selectedBrick.owner.BrickHeight);
            }
        }

        #endregion

        #region Properties

        public String PaddleImage
        {
            set
            {
                paddleAttribs.fileName = value;
                paddleImage = Image.FromFile(System.IO.Path.Combine(DataPaths.Paths["Textures"], value));
            }
        }
        public String BallImage
        {
            set
            {
                ballAttribs.fileName = value;
                ballImage = Image.FromFile(System.IO.Path.Combine(DataPaths.Paths["Textures"], value));
            }
        }
        public int PaddleX
        {
            set
            {
                paddleAttribs.startPosition.X = value;
                paddlePosition.X = TransformToScreenSpaceX(value) - (paddleAttribs.width / 2);
            }
        }
        public int PaddleY
        {
            set
            {
                paddleAttribs.startPosition.Y = value;
                paddlePosition.Y = TransformToScreenSpaceY(value) - (paddleAttribs.height / 2);
            }
        }
        public int PaddleWidth
        {
            set
            {
                paddleAttribs.width = value;
                paddlePosition.X = TransformToScreenSpaceX(paddleAttribs.startPosition.X) - (value / 2);
            }
        }
        public int PaddleHeight
        {
            set
            {
                paddleAttribs.height = value;
                paddlePosition.Y = TransformToScreenSpaceY(paddleAttribs.startPosition.Y) - (value / 2);
            }
        }

        #endregion
    }
}
