﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;

namespace LevelEditor
{
    class LevelParser
    {
        public static bool Parse(String file, Level lvl)
        {
            int startTime = System.Environment.TickCount;

            ResetStatus();

            lvl.Revert();

            XmlReaderSettings xmlSettings = new XmlReaderSettings();
            xmlSettings.IgnoreComments = true;
            xmlSettings.IgnoreWhitespace = true;

            try
            {
                XmlReader reader = XmlReader.Create(file, xmlSettings);
                XmlDocument doc = new XmlDocument();
                doc.Load(reader);

                if (doc.FirstChild.Name != "Level")
                {
                    lastError = "Level file '" + System.IO.Path.GetFileName(file) + "' is missing the root node <Level> ... </level>.";
                    return false;
                }

                XmlNode levelNodes = doc.FirstChild;

                System.Collections.IEnumerator nodes = levelNodes.GetEnumerator();
                uint count = 0;
                while (nodes.MoveNext()) count++;

                step = 100.0f / System.Math.Max(count, 1);

                bool result = TopLevelParser(levelNodes, lvl);

                reader.Close();

                loadTimeMS = System.Environment.TickCount - startTime;

                return result;
            }
            catch (System.IO.FileNotFoundException e)
            {
                lastError = "File '" + e.Message + "' could not be found.";
                return false;
            }
            catch (XmlException e)
            {
                lastError = "XML error in '" + System.IO.Path.GetFileName(file) + "' -> " + e.Message;
                return false;
            }
        }

        static public bool WriteLevel(String file, Level lvl)
        {
            int startTime = System.Environment.TickCount;

            ResetStatus();

            XmlDocument doc = new XmlDocument();
            XmlElement root = doc.CreateElement("Level");
            doc.AppendChild(root);

            WriteLevelAttributes(doc, root, lvl.levelAttribs);
            WritePaddleAttributes(doc, root, lvl.paddleAttribs);
            WriteBallAttributes(doc, root, lvl.ballAttribs);

            foreach(Level.BrickGroupAttributes group in lvl.brickGroups)
            {
                if (!WriteBrickGroup(doc, root, group))
                    return false;
            }

            doc.Save(System.IO.Path.Combine(DataPaths.Paths["Levels"], file));

            Percentage = 100.0f;

            loadTimeMS = System.Environment.TickCount - startTime;

            return true;
        }

        #region Level Parsing Status

        public delegate void PercentageChangeCallback(float lastPercentage, float newPercentage);

        public static PercentageChangeCallback percentageCallback = null;

        private static float percentage = 0.0f;
        public static float Percentage 
        {
            get
            {
                return percentage;
            }

            private set
            {
                if (percentageCallback != null)
                {
                    percentageCallback(percentage, value);
                }

                percentage = value;
            }
        }

        private static float step;

        public static int loadTimeMS { get; private set; }

        public static bool WasError() { return lastError != null; } 
        public static String lastError { get; private set; }

        private static List<String> warnings = new List<string>();
        public static List<String> Warnings() { return warnings; }

        private static void ResetStatus()
        {
            lastError = null;
            warnings.Clear();
            Percentage = 0.0f;
        }

        #endregion

        #region Level File Directives

        private enum LEVEL_FILE_DIRECTIVE
        {
	        // Top level level file directives
	        DIRECTIVE_LOAD = 0,
	        DIRECTIVE_BRICKGROUP,
	        DIRECTIVE_LEVEL_ATTRIBUTES,
	        DIRECTIVE_PADDLE_ATTRIBUTES,
	        DIRECTIVE_BALL_ATTRIBUTES,

	        // Level attribute directives
	        DIRECTIVE_BACKGROUND_COLOR,
	        DIRECTIVE_BOUNDS,

	        // BrickGroup directives
	        DIRECTIVE_BRICK_ATTRIBUTES,
	        DIRECTIVE_DISTROBUTION,

	        DIRECTIVE_UNKNOWN
        }

        // Top level level file directives
        private const String LOAD_STR               = "Load";
        private const String BRICKGROUP_STR         = "BrickGroup";
        private const String LEVEL_ATTRIBUTES_STR   = "LevelAttributes";
        private const String PADDLE_ATTRIBUTES_STR  = "PaddleAttributes";
        private const String BALL_ATTRIBUTES_STR    = "BallAttributes";

        // Level attribute directives
        private const String BACKGROUND_COLOR_STR   = "BackgroundColor";
        private const String BOUNDS_STR             = "Bounds";

        // BrickGroup directives
        private const String BRICK_ATTRIBUTES_STR   = "BrickAttributes";
        private const String DISTROBUTION_STR       = "Distrobution";

        // Distrobution strings
        private const String RANDOMDISTRO_STR       = "Random";
        private const String UNIFORMDISTRO_STR      = "Uniform";
        private const String POINTSDISTRO_STR       = "Points";
        private const String POINT_STR              = "Point";

        private static LEVEL_FILE_DIRECTIVE GetDirective(String str)
        {
            if (str == LOAD_STR)                return LEVEL_FILE_DIRECTIVE.DIRECTIVE_LOAD;
            if (str == BRICKGROUP_STR)          return LEVEL_FILE_DIRECTIVE.DIRECTIVE_BRICKGROUP;
            if (str == BACKGROUND_COLOR_STR)    return LEVEL_FILE_DIRECTIVE.DIRECTIVE_BACKGROUND_COLOR;
            if (str == BOUNDS_STR)              return LEVEL_FILE_DIRECTIVE.DIRECTIVE_BOUNDS;
            if (str == LEVEL_ATTRIBUTES_STR)    return LEVEL_FILE_DIRECTIVE.DIRECTIVE_LEVEL_ATTRIBUTES;
            if (str == BRICK_ATTRIBUTES_STR)    return LEVEL_FILE_DIRECTIVE.DIRECTIVE_BRICK_ATTRIBUTES;
            if (str == PADDLE_ATTRIBUTES_STR)   return LEVEL_FILE_DIRECTIVE.DIRECTIVE_PADDLE_ATTRIBUTES;
            if (str == BALL_ATTRIBUTES_STR)     return LEVEL_FILE_DIRECTIVE.DIRECTIVE_BALL_ATTRIBUTES;
            if (str == DISTROBUTION_STR)        return LEVEL_FILE_DIRECTIVE.DIRECTIVE_DISTROBUTION;

            return LEVEL_FILE_DIRECTIVE.DIRECTIVE_UNKNOWN;
        }

        private static Level.BrickGroupAttributes.DistrobutionType GetDistroType(String str)
        {
	        if(str == RANDOMDISTRO_STR)		return Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_RANDOM;
	        if(str == UNIFORMDISTRO_STR)	return Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNIFORM;
	        if(str == POINTSDISTRO_STR)		return Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_POINTS;

	        return Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNKNOWN;
        }

        private static String GetDistroString(Level.BrickGroupAttributes.DistrobutionType type)
        {
            switch (type)
            {
                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_RANDOM:
                    return RANDOMDISTRO_STR;
                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNIFORM:
                    return UNIFORMDISTRO_STR;
                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_POINTS:
                    return POINTSDISTRO_STR;
                default:
                    return null;
            };
        }

        #endregion

        #region Writers

        private static bool WriteLevelAttributes(XmlDocument doc, XmlElement root, Level.LevelAttributes att)
        {
            XmlElement levelAttribs = doc.CreateElement(LEVEL_ATTRIBUTES_STR);
            root.AppendChild(levelAttribs);

            XmlAttribute name = doc.CreateAttribute("name");
            name.Value = att.levelName;
            levelAttribs.Attributes.Append(name);

            XmlAttribute font = doc.CreateAttribute("font");
            font.Value = att.fontFile.Replace('\\', '/'); ;
            levelAttribs.Attributes.Append(font);

            XmlAttribute fontSize = doc.CreateAttribute("fontSize");
            fontSize.Value = Convert.ToString(att.fontSize);
            levelAttribs.Attributes.Append(fontSize);

            XmlElement bgColor = doc.CreateElement(BACKGROUND_COLOR_STR);
            levelAttribs.AppendChild(bgColor);

            XmlAttribute r = doc.CreateAttribute("r");
            r.Value = Convert.ToString(att.bgColor.R);
            bgColor.Attributes.Append(r);

            XmlAttribute g = doc.CreateAttribute("g");
            g.Value = Convert.ToString(att.bgColor.G);
            bgColor.Attributes.Append(g);

            XmlAttribute b = doc.CreateAttribute("b");
            b.Value = Convert.ToString(att.bgColor.B);
            bgColor.Attributes.Append(b);

            return true;
        }

        private static bool WriteBallAttributes(XmlDocument doc, XmlElement root, Level.BallAttributes att)
        {
            XmlElement ballAttribs = doc.CreateElement(BALL_ATTRIBUTES_STR);
            root.AppendChild(ballAttribs);

            XmlAttribute image = doc.CreateAttribute("image");
            image.Value = att.fileName.Replace('\\', '/'); ;
            ballAttribs.Attributes.Append(image);

            XmlAttribute radius = doc.CreateAttribute("radius");
            radius.Value = Convert.ToString(att.radius);
            ballAttribs.Attributes.Append(radius);

            XmlAttribute x = doc.CreateAttribute("x");
            x.Value = Convert.ToString(att.startPosition.X);
            ballAttribs.Attributes.Append(x);

            XmlAttribute y = doc.CreateAttribute("y");
            y.Value = Convert.ToString(att.startPosition.Y);
            ballAttribs.Attributes.Append(y);

            XmlAttribute velx = doc.CreateAttribute("velx");
            velx.Value = Convert.ToString(att.initVelocity.X);
            ballAttribs.Attributes.Append(velx);

            XmlAttribute vely = doc.CreateAttribute("vely");
            vely.Value = Convert.ToString(att.initVelocity.Y);
            ballAttribs.Attributes.Append(vely);

            return true;
        }

        private static bool WritePaddleAttributes(XmlDocument doc, XmlElement root, Level.PaddleAttributes att)
        {
            XmlElement paddleAttribs = doc.CreateElement(PADDLE_ATTRIBUTES_STR);
            root.AppendChild(paddleAttribs);

            XmlAttribute image = doc.CreateAttribute("image");
            image.Value = att.fileName.Replace('\\', '/');
            paddleAttribs.Attributes.Append(image);

            XmlAttribute shape = doc.CreateAttribute("shape");
            shape.Value = att.shapeFile.Replace('\\', '/');
            paddleAttribs.Attributes.Append(shape);

            XmlAttribute width = doc.CreateAttribute("width");
            width.Value = Convert.ToString(att.width);
            paddleAttribs.Attributes.Append(width);

            XmlAttribute height = doc.CreateAttribute("height");
            height.Value = Convert.ToString(att.height);
            paddleAttribs.Attributes.Append(height);

            XmlAttribute x = doc.CreateAttribute("x");
            x.Value = Convert.ToString(att.startPosition.X);
            paddleAttribs.Attributes.Append(x);

            XmlAttribute y = doc.CreateAttribute("y");
            y.Value = Convert.ToString(att.startPosition.Y);
            paddleAttribs.Attributes.Append(y);

            return true;
        }

        private static bool WriteBrickGroup(XmlDocument doc, XmlElement root, Level.BrickGroupAttributes group)
        {
            XmlElement brickGroup = doc.CreateElement(BRICKGROUP_STR);
            root.AppendChild(brickGroup);

            XmlAttribute name = doc.CreateAttribute("name");
            name.Value = group.name;
            brickGroup.Attributes.Append(name);

            XmlElement brickAttrib = doc.CreateElement(BRICK_ATTRIBUTES_STR);
            brickGroup.AppendChild(brickAttrib);

            XmlAttribute hp = doc.CreateAttribute("hp");
            hp.Value = Convert.ToString(group.brickHP);
            brickAttrib.Attributes.Append(hp);

            XmlAttribute pointsPerHit = doc.CreateAttribute("pointsPerHit");
            pointsPerHit.Value = Convert.ToString(group.pointsPerHit);
            brickAttrib.Attributes.Append(pointsPerHit);

            XmlAttribute pointsPerKill = doc.CreateAttribute("pointsPerKill");
            pointsPerKill.Value = Convert.ToString(group.pointsPerKill);
            brickAttrib.Attributes.Append(pointsPerKill);

            XmlAttribute width = doc.CreateAttribute("width");
            width.Value = Convert.ToString(group.brickWidth);
            brickAttrib.Attributes.Append(width);

            XmlAttribute height = doc.CreateAttribute("height");
            height.Value = Convert.ToString(group.brickHeight);
            brickAttrib.Attributes.Append(height);

            uint count = 0;
            String lastString = null;
            foreach (String s in group.imageFiles)
            {
                if (lastString == null || s != lastString)
                {
                    XmlAttribute a = doc.CreateAttribute("image" + count);
                    a.Value = s.Replace('\\', '/'); ;
                    brickAttrib.Attributes.Append(a);
                }

                lastString = s;
                count++;
            }

            XmlElement distro = doc.CreateElement(DISTROBUTION_STR);
            brickGroup.AppendChild(distro);

            XmlAttribute type = doc.CreateAttribute("type");
            String distroStr = GetDistroString(group.distType);
            if (distroStr == null)
            {
                lastError = "The attribute 'type' for the directive '" + DISTROBUTION_STR + "' for the BrickGroup named '"
                    + group.name + "' is not valid";
                return false;
            }
            else
            {
                type.Value = distroStr;
                distro.Attributes.Append(type);
            }

            switch (group.distType)
            {
                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_POINTS:
                    WritePointDistrobution(doc, distro, group);
                    break;

                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_RANDOM:
                    WriteRandomDistrobution(doc, distro, group);
                    break;

                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNIFORM:
                    WriteUniformDistrobution(doc, distro, group);
                    break;
            };


            return true;
        }

        private static void WritePointDistrobution(XmlDocument doc, XmlElement dist, Level.BrickGroupAttributes group)
        {
            foreach (System.Drawing.Point p in group.points)
            {
                XmlAttribute x = doc.CreateAttribute("x");
                x.Value = Convert.ToString(p.X);

                XmlAttribute y = doc.CreateAttribute("y");
                y.Value = Convert.ToString(p.Y);

                XmlElement point = doc.CreateElement("Point");
                point.Attributes.Append(x);
                point.Attributes.Append(y);

                dist.AppendChild(point);
            }
        }

        private static void WriteRandomDistrobution(XmlDocument doc, XmlElement dist, Level.BrickGroupAttributes group)
        {

        }

        private static void WriteUniformDistrobution(XmlDocument doc, XmlElement dist, Level.BrickGroupAttributes group)
        {

        }

        #endregion

        #region Parsers

        private static bool TopLevelParser(XmlNode nodes, Level lvl)
        {
            foreach (XmlNode n in nodes)
            {
                directiveParser func;
                if (parsers.TryGetValue(n.Name, out func))
                {
                    if (!func(n, lvl))
                        return false;
                }
                else
                {
                    lastError = "Level file command '" + n.Name + "' can not occur at the top level.";
                    return false;
                }

                if(n.Name != LOAD_STR)
                    Percentage += step;
            }

            return true;
        }

        private static bool ParseLoad(XmlNode n, Level lvl)
        {
            String fileName = null;

            foreach (XmlAttribute a in n.Attributes)
            {
                if (a.Name == "file")
                {
                    if (fileName == null)
                        fileName = a.Value;
                    else
                        warnings.Add("'file' attribute has already been specified in '" + LOAD_STR + "' command.");
                }
                else
                {
                    warnings.Add("Unknown attribute '" + a.Name + "' found in '" + LOAD_STR + "' command.");
                }
            }

            try
            {
                XmlReaderSettings xmlSettings = new XmlReaderSettings();
                xmlSettings.IgnoreComments = true;
                xmlSettings.IgnoreWhitespace = true;

                String f = System.IO.Path.Combine(DataPaths.Paths["Levels"], fileName);

                XmlReader reader = XmlReader.Create(System.IO.Path.Combine(DataPaths.Paths["Levels"], fileName), xmlSettings);
                XmlDocument doc = new XmlDocument();
                doc.Load(reader);

                if (doc.FirstChild.Name != "Level")
                {
                    lastError = "Level file '" + System.IO.Path.GetFileName(fileName) 
                        + "' is missing the root node <Level> ... </level>.";
                    return false;
                }

                float oldStep = step;

                XmlNode levelNodes = doc.FirstChild;
                System.Collections.IEnumerator nodes = levelNodes.GetEnumerator();

                uint count = 0;
                while (nodes.MoveNext()) count++;

                step = step / System.Math.Max(count, 1);

                if (!TopLevelParser(levelNodes, lvl))
                {
                    lastError = "XML file does not define a valid level.";
                    return false;
                }

                step = oldStep;

                reader.Close();

                return true;
            }
            catch (System.IO.FileNotFoundException e)
            {
                lastError = "File '" + e.FileName + "' could not be found.";
                return false;
            }
            catch (XmlException e)
            {
                lastError = "XML error in '" + System.IO.Path.GetFileName(fileName) + "' -> " + e.Message;
                return false;
            }
        }

        private static bool ParseLevelAttributes(XmlNode n, Level lvl)
        {
            foreach (XmlAttribute a in n.Attributes)
            {
                if (a.Name == "name")
                {
                    lvl.levelAttribs.levelName = a.Value;
                }
                else if (a.Name == "font")
                {
                    lvl.levelAttribs.fontFile = a.Value;
                }
                else if (a.Name == "fontSize")
                {
                    try
                    {
                        lvl.levelAttribs.fontSize = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.levelAttribs.fontSize = 12;
                        warnings.Add("'fontSize' attribute of '" + LEVEL_ATTRIBUTES_STR 
                            + "' is not a properly formatted int -> " + e.Message );
                    }
                }
                else
                {
                    warnings.Add("'" + a.Name + "' is not a supported attribute for the '" 
                        + LEVEL_ATTRIBUTES_STR + "' directive.");
                }
            }

            foreach (XmlNode node in n)
            {
                if (node.Name == BACKGROUND_COLOR_STR)
                {
                    int r, g, b;
                    try
                    {
                        if (!ParseInt(node.Attributes["r"].Value, out r) || r < 0 || r > 255)
                        {
                            warnings.Add("'r' attribute of the '" + BACKGROUND_COLOR_STR
                                + "' directive is not a properly formatted byte in the range [0, 255].");
                        }
                    }
                    catch (ArgumentOutOfRangeException)
                    {
                        r = 255;
                        warnings.Add("'" + BACKGROUND_COLOR_STR + "' directive is missing an 'r' attribute.");
                    }

                    try
                    {
                        if (!ParseInt(node.Attributes["g"].Value, out g) || g < 0 || g > 255)
                        {
                            warnings.Add("'g' attribute of the '" + BACKGROUND_COLOR_STR
                                + "' directive is not a properly formatted byte in the range [0, 255].");
                        }
                    }
                    catch (ArgumentOutOfRangeException)
                    {
                        g = 255;
                        warnings.Add("'" + BACKGROUND_COLOR_STR + "' directive is missing an 'g' attribute.");
                    }

                    try
                    {
                        if (!ParseInt(node.Attributes["b"].Value, out b) || b < 0 || b > 255)
                        {
                            warnings.Add("'b' attribute of the '" + BACKGROUND_COLOR_STR
                                + "' directive is not a properly formatted byte in the range [0, 255].");
                        }
                    }
                    catch (ArgumentOutOfRangeException)
                    {
                        b = 255;
                        warnings.Add("'" + BACKGROUND_COLOR_STR + "' directive is missing an 'b' attribute.");
                    }

                    lvl.levelAttribs.bgColor = System.Drawing.Color.FromArgb(r, g, b);
                }
                else if (node.Name == BOUNDS_STR)
                {
                    warnings.Add("The '" + BOUNDS_STR + "' directive is no longer used and is ignored.");
                }
                else
                {
                    warnings.Add("'" + node.Name + "' is not a supported directive under the '"
                        + LEVEL_ATTRIBUTES_STR + "' directice the.");
                }
            }

            return true;
        }

        private static bool ParseBallAttributes(XmlNode n, Level lvl)
        {
            foreach (XmlAttribute a in n.Attributes)
            {
                if (a.Name == "image")
                {
                    lvl.ballAttribs.fileName = a.Value;
                }
                else if (a.Name == "radius")
                {
                    try
                    {
                        lvl.ballAttribs.radius = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.levelAttribs.fontSize = 12;
                        warnings.Add("'radius' attribute of '" + BALL_ATTRIBUTES_STR + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "x")
                {
                    try
                    {
                        lvl.ballAttribs.startPosition.X = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.ballAttribs.startPosition.X = 0;
                        warnings.Add("'x' attribute of '" + BALL_ATTRIBUTES_STR + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "y")
                {
                    try
                    {
                        lvl.ballAttribs.startPosition.Y = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.ballAttribs.startPosition.Y = 0;
                        warnings.Add("'y' attribute of '" + BALL_ATTRIBUTES_STR + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "velx")
                {
                    try
                    {
                        lvl.ballAttribs.initVelocity.X = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.ballAttribs.initVelocity.X = 0;
                        warnings.Add("'velx' attribute of '" + BALL_ATTRIBUTES_STR 
                            + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "vely")
                {
                    try
                    {
                        lvl.ballAttribs.initVelocity.Y = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.ballAttribs.initVelocity.Y = 0;
                        warnings.Add("'vely' attribute of '" + BALL_ATTRIBUTES_STR 
                            + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else
                {
                    warnings.Add("'" + a.Name + "' is not a supported attribute for the '" 
                        + BALL_ATTRIBUTES_STR + "' directive.");
                }
            }

            return true;
        }

        private static bool ParsePaddleAttributes(XmlNode n, Level lvl)
        {
            foreach (XmlAttribute a in n.Attributes)
            {
                if (a.Name == "image")
                {
                    lvl.paddleAttribs.fileName = a.Value;
                }
                if (a.Name == "shape")
                {
                    lvl.paddleAttribs.shapeFile = a.Value;
                }
                else if (a.Name == "width")
                {
                    try
                    {
                        lvl.paddleAttribs.width = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.paddleAttribs.width = 64;
                        warnings.Add("'width' attribute of '" + PADDLE_ATTRIBUTES_STR
                            + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "height")
                {
                    try
                    {
                        lvl.paddleAttribs.height = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.paddleAttribs.width = 64;
                        warnings.Add("'height' attribute of '" + PADDLE_ATTRIBUTES_STR
                            + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "x")
                {
                    try
                    {
                        lvl.paddleAttribs.startPosition.X = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.paddleAttribs.startPosition.X = 0;
                        warnings.Add("'x' attribute of '" + PADDLE_ATTRIBUTES_STR
                            + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else if (a.Name == "y")
                {
                    try
                    {
                        lvl.paddleAttribs.startPosition.Y = int.Parse(a.Value);
                    }
                    catch (System.FormatException e)
                    {
                        lvl.paddleAttribs.startPosition.Y = 0;
                        warnings.Add("'y' attribute of '" + PADDLE_ATTRIBUTES_STR
                            + "' is not a properly formatted int -> " + e.Message);
                    }
                }
                else
                {
                    warnings.Add("'" + a.Name + "' is not a supported attribute for the '"
                        + PADDLE_ATTRIBUTES_STR + "' directive.");
                }
            }

            return true;
        }

        private static bool ParseBrickGroup(XmlNode n, Level lvl)
        {
            Level.BrickGroupAttributes brickGroup = new Level.BrickGroupAttributes();

            foreach (XmlAttribute a in n.Attributes)
            {
                if (a.Name == "name")
                {
                    brickGroup.name = a.Value;
                }
                else
                {
                    warnings.Add("'" + a.Name + "' is not a supported attribute for the '"
                        + BRICKGROUP_STR + "' directive.");
                }
            }

            foreach (XmlNode node in n)
            {
                if (node.Name == BRICK_ATTRIBUTES_STR)
                {
                    if (!ParseBrickAttributes(node, brickGroup))
                        return false;
                }
                else if (node.Name == DISTROBUTION_STR)
                {
                    if (brickGroup.distType != Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNKNOWN)
                    {
                        warnings.Add("Multiple brick Distributions have be specified under the BrickGroup '" 
                            + brickGroup.name + "'.  Ignoring redundent distrobution");
                    }
                    else
                    {
                        ParseDistrobution(node, brickGroup);
                    }
                }
            }

            lvl.brickGroups.Add(brickGroup);

            return true;
        }

        private static bool ParseBrickAttributes(XmlNode n, Level.BrickGroupAttributes brickGroup)
        {
            foreach (XmlAttribute a in n.Attributes)
            {
                if (a.Name == "hp")
                {
                    if (!ParseInt(a.Value, out brickGroup.brickHP))
                    {
                        brickGroup.brickHP = 0;
                        warnings.Add("'hp' attribute of '" + BRICK_ATTRIBUTES_STR
                            + "' is not a properly formatted int.");
                    }
                }
                else if (a.Name == "pointsPerHit")
                {
                    if (!ParseInt(a.Value, out brickGroup.pointsPerHit))
                    {
                        brickGroup.pointsPerHit = 0;
                        warnings.Add("'pointsPerHit' attribute of '" + BRICK_ATTRIBUTES_STR
                            + "' is not a properly formatted int.");
                    }
                }
                else if (a.Name == "pointsPerKill")
                {
                    if (!ParseInt(a.Value, out brickGroup.pointsPerKill))
                    {
                        brickGroup.pointsPerKill = 0;
                        warnings.Add("'pointsPerKill' attribute of '" + BRICK_ATTRIBUTES_STR
                            + "' is not a properly formatted int.");
                    }
                }
                else if (a.Name == "width")
                {
                    if (!ParseInt(a.Value, out brickGroup.brickWidth))
                    {
                        brickGroup.brickWidth = 0;
                        warnings.Add("'width' attribute of '" + BRICK_ATTRIBUTES_STR
                            + "' is not a properly formatted int.");
                    }
                }
                else if (a.Name == "height")
                {
                    if (!ParseInt(a.Value, out brickGroup.brickHeight))
                    {
                        brickGroup.brickHeight = 0;
                        warnings.Add("'height' attribute of '" + BRICK_ATTRIBUTES_STR
                            + "' is not a properly formatted int.");
                    }
                }
            }

            for (int i = 0; i < brickGroup.brickHP; i++)
            {
                String imName = "image" + i;
                XmlAttribute a = n.Attributes[imName];
                if (a == null)
                {
                    if (brickGroup.imageFiles.Count > 0)
                        brickGroup.imageFiles.Add(brickGroup.imageFiles[brickGroup.imageFiles.Count - 1]);
                    else
                        throw new OverflowException();
                }
                else
                {
                    brickGroup.imageFiles.Add(a.Value);
                }
            }

            return true;
        }

        private static bool ParseDistrobution(XmlNode n, Level.BrickGroupAttributes brickGroup)
        {
            Level.BrickGroupAttributes.DistrobutionType dist = Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNKNOWN;

            try
            {
                dist = GetDistroType(n.Attributes["type"].Value);
            }
            catch (ArithmeticException)
            {
                warnings.Add("'" + DISTROBUTION_STR 
                    + "' directive is missing the 'type' attribute under the BrickGroup '" 
                    + brickGroup.name + "'");
                return true; // Not a fatal error
            }

            brickGroup.distType = dist;

            switch (dist)
            {
                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_POINTS:
                    return ParsePointDistrobution(n, brickGroup);

                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_RANDOM:
                    return ParseRandomDistrobution(n, brickGroup);

                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNIFORM:
                    return ParseUniformDistrobution(n, brickGroup);

                case Level.BrickGroupAttributes.DistrobutionType.DISTROTYPE_UNKNOWN:
                default:
                    warnings.Add("'" + DISTROBUTION_STR 
                        + "' directive 'type' attribute names an unknown type under the BrickGroup '" 
                        + brickGroup.name + "'");
                    return true; // not a fatal error
            };
        }

        private static bool ParsePointDistrobution(XmlNode n, Level.BrickGroupAttributes brickGroup)
        {
            foreach (XmlNode node in n)
            {
                if (node.Name == "Point")
                {
                    int x, y;
                    ParsePoint(node, out x, out y);
                    brickGroup.points.Add(new System.Drawing.Point(x, y));
                }
                else
                {
                    warnings.Add("'" + node.Name + "' is not a supported directive for the in a Point Distrobution");
                }
            }

            return true;
        }

        private static bool ParseRandomDistrobution(XmlNode n, Level.BrickGroupAttributes brickGroup)
        {
            return true;
        }

        private static bool ParseUniformDistrobution(XmlNode n, Level.BrickGroupAttributes brickGroup)
        {
            return true;
        }

        private static bool ParsePoint(XmlNode n, out int x, out int y)
        {
            try
            {
                if (!ParseInt(n.Attributes["x"].Value, out x))
                {
                    warnings.Add("'x' attribute of the 'Point' directive is not a properly formatted int.");
                }
            }
            catch (ArgumentOutOfRangeException)
            {
                x = 0;
                warnings.Add("'Point' directive is missing an 'x' attribute.");
            }

            try
            {
                if (!ParseInt(n.Attributes["y"].Value, out y))
                {
                    warnings.Add("'y' attribute of the 'Point' directive is not a properly formatted int.");
                }
            }
            catch (ArgumentOutOfRangeException)
            {
                y = 0;
                warnings.Add("'Point' directive is missing a 'y' attribute.");
            }

            return true;
        }

        private static bool ParseInt(String str, out int val)
        {
            try
            {
                val = int.Parse(str);
                return true;
            }
            catch (System.FormatException)
            {
                val = 0;
                return false;
            }
        }

        #endregion

        #region Parser Dictionary

        private delegate bool directiveParser(XmlNode n, Level lvl);

        private static Dictionary<String, directiveParser> parsers = new Dictionary<String, directiveParser>();

        static LevelParser()
        {
            parsers.Add(LOAD_STR, ParseLoad);
            parsers.Add(LEVEL_ATTRIBUTES_STR, ParseLevelAttributes);
            parsers.Add(BRICKGROUP_STR, ParseBrickGroup);
            parsers.Add(BALL_ATTRIBUTES_STR, ParseBallAttributes);
            parsers.Add(PADDLE_ATTRIBUTES_STR, ParsePaddleAttributes);
        }

        #endregion
    }
}
