﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace LevelEditor
{
    public partial class MainWindow : Form
    {
        public static String WindowName = "BrickOut |ReDuX| Level Editor - ";

        public MainWindow()
        {
            MainWindow.window = this;

            InitializeComponent();
        }

        private void MainWindow_Load(object sender, EventArgs e)
        {
            Text = WindowName + currentLevelFile;

            ConsoleWriteLine(Text + " started " + DateTime.Now + "...\n");
            DataPaths.Load("DataPaths.xml");

            OpenLevelFileDialog.InitialDirectory = DataPaths.Paths["Levels"];
            SaveLevelFileDialog.InitialDirectory = DataPaths.Paths["Levels"];

            List<String> imageExt = new List<string>();
            imageExt.Add(".png");
            imageExt.Add(".bmp");
            imageExt.Add(".dds");
            imageExt.Add(".tga");
            imageExt.Add(".jpg");
            imageExt.Add(".jpeg");

            imageSelect.Text = "Texture Select...";
            imageSelect.SetDataPath(DataPaths.Paths["Textures"], imageExt);

            List<String> fontExt = new List<string>();
            fontExt.Add(".ttf");

            fontSelect.Text = "Font Select...";
            fontSelect.SetDataPath(DataPaths.Paths["Fonts"], fontExt);

            List<String> shapeExt = new List<string>();
            shapeExt.Add(".shape");

            shapeSelect.Text = "Shape Select...";
            shapeSelect.SetDataPath(DataPaths.Paths["Shapes"], shapeExt);

            // Default grid setup
            grid.ScreenWidth = Level.Width;
            grid.ScreenHeight = Level.Height;

            CellWidthNumericUpDown.Value = 64;
            CellHeightNumericUpDown.Value = 16;

            RefreshEditor(level);
        }

        public enum MouseTool
        {
            Select = 0,
            Move,
            Placment,
            Removal,

            ToolCount
        };

        #region Console IO

        private void ConsoleInputEnter()
        {
            if (ConsoleInput.Text.Length > 0)
            {
                ConsoleWriteLine(ConsoleInput.Text);
                ConsoleInput.Clear();
            }
        }

        public static void ConsoleWrite(String txt)
        {
            if (MainWindow.window.Console.InvokeRequired)
            {
                MainWindow.window.Console.Invoke(new ConsoleWriteCallback(ConsoleWrite), new Object[] { txt });
            }
            else
            {
                MainWindow.window.Console.AppendText(txt);
                MainWindow.window.Console.ScrollToCaret();
            }
        }

        public static void ConsoleWriteLine(String txt)
        {
            if (MainWindow.window.Console.InvokeRequired)
            {
                MainWindow.window.Console.Invoke(new ConsoleWriteCallback(ConsoleWriteLine), new Object[] { txt });
            }
            else
            {
                MainWindow.window.Console.AppendText(txt + "\n");
                MainWindow.window.Console.ScrollToCaret();
            }
        }

        #endregion

        #region Game Launcher Stuff

        public static void LaunchGame(String args)
        {
            if (gameProcess != null)
            {
                ConsoleWriteLine("ERROR: Only one instance of BrickOut |ReDuX| can be active at once.");
                return;
            }

            gameProcess = new System.Diagnostics.Process();
            gameProcess.StartInfo.FileName = System.IO.Path.Combine(DataPaths.Paths["Bin"], "BrickOut ReDuX.exe");
            gameProcess.StartInfo.Arguments = args;

            gameProcess.Exited += new EventHandler(GameProcessExitedEventHandler);
            gameProcess.OutputDataReceived += new System.Diagnostics.DataReceivedEventHandler(GameProcessOutputDataReceivedEventHandler);
            gameProcess.ErrorDataReceived += new System.Diagnostics.DataReceivedEventHandler(GameProcessErrorDataReceivedEventHandler);

            gameProcess.EnableRaisingEvents = true;
            gameProcess.StartInfo.UseShellExecute = false;
            gameProcess.StartInfo.RedirectStandardOutput = true;
            gameProcess.StartInfo.RedirectStandardError = true;
            gameProcess.Start();

            gameProcess.BeginOutputReadLine();
            gameProcess.BeginErrorReadLine();

            ConsoleWriteLine("BrickOut |ReDuX| launched at " + System.DateTime.Now + " with arguments : " + args);
        }

        static void GameProcessErrorDataReceivedEventHandler(object sender, System.Diagnostics.DataReceivedEventArgs e)
        {
            ConsoleWriteLine("stderr -> " + e.Data);
        }

        private static void GameProcessOutputDataReceivedEventHandler(object sender, System.Diagnostics.DataReceivedEventArgs e)
        {
            ConsoleWriteLine("stdout -> " + e.Data);
        }

        private static void GameProcessExitedEventHandler(object sender, EventArgs e)
        {

            ConsoleWriteLine("...BrickOut |ReDuX| exited with code " + gameProcess.ExitCode 
                + " at time " + gameProcess.ExitTime + ".");

            gameProcess = null;
        }

        #endregion

        #region Level Maintenance

        private void ReDrawLevel()
        {
            BufferedGraphicsContext context = BufferedGraphicsManager.Current;

            Graphics gfx = LevelViewer.CreateGraphics();
            BufferedGraphics g = context.Allocate(gfx, LevelViewer.ClientRectangle);

            level.Draw(g.Graphics);

            g.Render(gfx);
        }

        private void PromptAndOpenLevelFile()
        {
            if (OpenLevelFileDialog.ShowDialog() != DialogResult.OK)
                return;

            String filePath = OpenLevelFileDialog.FileName;
            LoadLevel(filePath);
            levelLoaded = true;
        }

        private bool PromptAndSaveLevelFile()
        {
            SaveLevelFileDialog.FileName = Path.GetFileName(currentLevelFile);
            if (SaveLevelFileDialog.ShowDialog() != DialogResult.OK)
                return false;

            String filePath = SaveLevelFileDialog.FileName;
            SaveLevel(filePath);
            levelLoaded = true;
            return true;
        }

        private void Save()
        {
            if (!levelLoaded)
                PromptAndSaveLevelFile();
            else
                SaveLevel(currentLevelFile);
        }

        private void Open()
        {
            if (PromptDataChanged())
            {
                PromptAndOpenLevelFile();
                ReDrawLevel();
            }
        }

        private void LevelSaveOrLoadPercentageCallback(float lastPercentage, float newPercentage)
        {
            StatusStripProgressBar.Increment((int)newPercentage);
            System.Threading.Thread.Sleep(50);
        }

        private void LoadLevel(String filePath)
        {
            String fileName = System.IO.Path.GetFileName(filePath);

            String oldText = StatusStripLabel.Text;
            StatusStripLabel.Text = "Opening " + fileName + "...";

            StatusStripProgressBar.Visible = true;
            StatusStripProgressBar.Value = 0;

            level.Revert();

            LevelParser.percentageCallback = LevelSaveOrLoadPercentageCallback;
            LevelParser.Parse(filePath, level);
            LevelParser.percentageCallback = null;

            if (LevelParser.WasError())
            {
                ConsoleWriteLine("Level file '" + filePath + "' could not be loaded.");
                ConsoleWriteLine("\tERROR: " + LevelParser.lastError);

                currentLevelFile = null;
                SaveToolStripButton.ToolTipText = "Save";
                Text = "BrickOut |ReDuX| - LevelEditor";
            }
            else
            {
                System.Collections.IEnumerator warnings = LevelParser.Warnings().GetEnumerator();
                uint count = 0;
                while (warnings.MoveNext()) count++;

                level.Init();

                ConsoleWriteLine("Level file '" + fileName + "' loaded in " + LevelParser.loadTimeMS / 1000.0f
                    + " seconds with " + count + " warnings");

                uint index = 1;
                foreach (String str in LevelParser.Warnings())
                {
                    ConsoleWriteLine("\t" + index + ". " + str);
                    index++;
                }

                RefreshEditor(level);

                currentLevelFile = filePath;
                SaveToolStripButton.ToolTipText = "Save " + fileName;
                Text = "BrickOut |ReDuX| - " + fileName;
            }

            StatusStripProgressBar.Visible = false;
            StatusStripLabel.Text = oldText;
        }

        private void SaveLevel(String filePath)
        {
            String fileName = System.IO.Path.GetFileName(filePath);

            String oldText = StatusStripLabel.Text;
            StatusStripLabel.Text = "Saving level file " + fileName + " ...";

            StatusStripProgressBar.Visible = true;
            StatusStripProgressBar.Value = 0;

            LevelParser.percentageCallback = LevelSaveOrLoadPercentageCallback;
            LevelParser.WriteLevel(filePath, level);
            LevelParser.percentageCallback = null;

            if (LevelParser.WasError())
            {
                ConsoleWriteLine("ERROR: '" + filePath + "' could not be saved -> " + LevelParser.lastError);
            }
            else
            {
                System.Collections.IEnumerator warnings = LevelParser.Warnings().GetEnumerator();
                uint count = 0;
                while (warnings.MoveNext()) count++;

                ConsoleWriteLine("Level file '" + fileName + "' saved in " + LevelParser.loadTimeMS / 1000.0f
                    + " seconds with " + count + " warnings");

                uint index = 1;
                foreach (String str in LevelParser.Warnings())
                {
                    ConsoleWriteLine("\t" + index + ". " + str);
                    index++;
                }

                currentLevelFile = filePath;
                SaveToolStripButton.ToolTipText = "Save " + fileName;
                Text = "BrickOut |ReDuX| - " + fileName;
            }

            StatusStripProgressBar.Visible = false;
            StatusStripLabel.Text = oldText;
        }

        private bool PromptDataChanged()
        {
            if (dataChanged)
            {
                DialogResult res = MessageBox.Show(
                    "Do you want to save chages to " + Path.GetFileName(currentLevelFile) + "?",
                    "Save Changes",
                    MessageBoxButtons.YesNoCancel,
                    MessageBoxIcon.Question,
                    MessageBoxDefaultButton.Button3);

                if (res == DialogResult.Yes)
                {
                    if (levelLoaded)
                    {
                        SaveLevel(currentLevelFile);
                    }
                    else
                    {
                        if (!PromptAndSaveLevelFile())
                            return false;
                    }
                }

                if (res != DialogResult.Cancel)
                {
                    DataChanged = false;
                    return true;
                }

                return false;
            }

            return true;
        }

        #endregion

        #region Level Editor Stuff

        private void SetTool(MouseTool t)
        {
            currentTool = t;
            switch (t)
            {
                case MouseTool.Select:
                    LevelViewer.Cursor = Cursors.Arrow;
                    break;
                case MouseTool.Move:
                    LevelViewer.Cursor = Cursors.SizeAll;
                    break;
                case MouseTool.Placment:
                    LevelViewer.Cursor = Cursors.Cross;
                    break;
                case MouseTool.Removal:
                    LevelViewer.Cursor = Cursors.No;
                    break;
            }
        }

        private void RefreshEditor(Level lvl)
        {
            ignoreDataChange = true;

            FontTextBox.Text = lvl.levelAttribs.fontFile;
            FontSizeNumericUpDown.Value = lvl.levelAttribs.fontSize;

            LevelNameTextBox.Text = lvl.levelAttribs.levelName;

            BrickGroupsListBox.DataSource = lvl.brickCollections;

            PaddleImageTextBox.Text = level.paddleAttribs.fileName;
            PaddleShapeTextBox.Text = level.paddleAttribs.shapeFile;
            PaddleWidthNumericUpDown.Value = level.paddleAttribs.width;
            PaddleHeightNumericUpDown.Value = level.paddleAttribs.height;
            PaddleXNumericUpDown.Value = level.paddleAttribs.startPosition.X;
            PaddleYNumericUpDown.Value = level.paddleAttribs.startPosition.Y;

            BallImageTextBox.Text = level.ballAttribs.fileName;
            BallRadiusNumericUpDown.Value = level.ballAttribs.radius;

            ignoreDataChange = false;
        }

        private Level.BrickCollection GetSelectedBrickGroup()
        {
            int index = BrickGroupsListBox.SelectedIndex;
            if (index >= 0)
                return (Level.BrickCollection)BrickGroupsListBox.Items[index];
            return null;
        }

        private void ExpandImageFileList(List<String> images, int amt)
        {
            int num = amt - images.Count;

            if (num > 0)
            {
                int start = images.Count;
                for (int i = 0; i < num; i++)
                {
                    int last = start + i - 1;
                    if (last >= 0)
                        images.Add(images[start + i - 1]);
                    else
                        images.Add("");
                }
            }
            else
            {
                num = -num;
                for (int i = 0; i < num; i++)
                {
                    images.RemoveAt(images.Count - 1);
                }
            }
        }

        private void ExpandImageList(int amt)
        {
            ImageCheckedListBox.BeginUpdate();

            int num = amt - ImageCheckedListBox.Items.Count;
            if (num > 0)
            {
                int start = ImageCheckedListBox.Items.Count;
                for (int i = 0; i < num; i++)
                {
                    ImageCheckedListBoxItem item = new ImageCheckedListBoxItem();
                    item.listName = "Image " + Convert.ToString(start + i);
                    item.index = start + i;
                    ImageCheckedListBox.Items.Add(item);
                }
            }
            else
            {
                num = -num;
                for (int i = 0; i < num; i++)
                {
                    ImageCheckedListBox.Items.RemoveAt(ImageCheckedListBox.Items.Count - 1);
                }
            }

            ImageCheckedListBox.EndUpdate();
        }

        private void UpdateImageFileList(List<String> images)
        {
            ImageCheckedListBox.ClearSelected();

            ignoreItemCheckedEvent = true;

            int len = images.Count;
            for (int i = 0; i < len; i++)
            {
                ImageCheckedListBoxItem item = (ImageCheckedListBoxItem)ImageCheckedListBox.Items[i];
                item.imagePath = images[i];
                item.imageName = Path.GetFileName(images[i]);

                int last = i - 1;
                if (last < 0 || images[i] != images[last])
                {
                    ImageCheckedListBox.SetItemChecked(i, true);
                }
            }

            ignoreItemCheckedEvent = false;
        }

        #endregion

        #region Events

        //
        // Level Viewer
        //
        private void LevelViewer_Paint(object sender, PaintEventArgs e)
        {
            ReDrawLevel();
        }

        private void LevelViewer_MouseClick(object sender, MouseEventArgs e)
        {

        }

        private void LevelViewer_MouseDown(object sender, MouseEventArgs e)
        {
            mouseDown = true;

            Level.Brick b = level.Pick(e.Location);

            if (currentTool == MouseTool.Placment)
            {
                Level.BrickCollection g = GetSelectedBrickGroup();
                if (g != null)
                {
                    Point pos = e.Location;
                    if (snapToGrid)
                        grid.SnapTo(ref pos);

                    Level.Brick newBrick = g.AddBrickScreenPosition(pos);
                    PointsCountLabel.Text = Convert.ToString(g.bricks.Count);

                    PointsListBox.SelectedIndex = newBrick.attribIndex;
                    level.SelectBrick(newBrick);

                    DataChanged = true;
                }
            }
            else if (currentTool == MouseTool.Select || currentTool == MouseTool.Move)
            {
                dragBrick = b;

                if (b != null)
                {
                    level.SelectBrick(b);

                    BrickGroupsListBox.SelectedItem = b.owner;
                    PointsListBox.SelectedIndex = b.attribIndex;
                    PropertiesPane.SelectedTab = PropertiesPane.TabPages["BrickGroupsTab"];
                }
                else
                {
                    PointsListBox.SelectedIndex = -1;
                    level.SelectBrick(null);
                    BrickGroupsListBox.SelectedItem = null;
                }
            }
            else if (currentTool == MouseTool.Removal)
            {
                if (b != null)
                {
                    Level.BrickCollection g = b.owner;

                    g.RemoveBrick(b);
                    PointsCountLabel.Text = Convert.ToString(g.bricks.Count);

                    PointsListBox.SelectedIndex = -1;
                    level.SelectBrick(null);
                    level.SelectCollection(g);

                    DataChanged = true;
                }
            }

            ReDrawLevel();
        }

        private void LevelViewer_MouseEnter(object sender, EventArgs e)
        {

        }

        private void LevelViewer_MouseHover(object sender, EventArgs e)
        {

        }

        private void LevelViewer_MouseLeave(object sender, EventArgs e)
        {
            mouseDown = false;
            dragBrick = null;
        }

        private void LevelViewer_MouseMove(object sender, MouseEventArgs e)
        {
            if (dragBrick != null && currentTool == MouseTool.Move && mouseDown)
            {
                Point pos = e.Location;
                if (snapToGrid)
                    grid.SnapTo(ref pos);

                dragBrick.MoveTo(pos);
                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void LevelViewer_MouseUp(object sender, MouseEventArgs e)
        {
            if (dragBrick != null)
            {
                ignoreDataChange = true;
                dragBrick.owner.bricks.ResetItem(dragBrick.attribIndex);
                PointXNumericUpDown.Value = dragBrick.position.X;
                PointYNumericUpDown.Value = dragBrick.position.Y;
                ignoreDataChange = false;
            }

            mouseDown = false;
            dragBrick = null;
        }

        //
        // Menu Items
        //
        private void FileQuitMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void SaveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            PromptAndSaveLevelFile();
        }

        private void OpenToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void CloseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            currentLevelFile = null;
            level.Revert();
            ReDrawLevel();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Save();
        }

        //
        // Tool Bar buttons
        //
        private void RunToolStripButton_Click(object sender, EventArgs e)
        {
            if (currentLevelFile == null)
            {
                ConsoleWriteLine("WARNING: Launching the game without an active map loaded.");
            }
            else
            {
                SaveLevel(currentLevelFile);
            }

            LaunchGame("-level " + System.IO.Path.GetFileName(currentLevelFile));
        }

        private void SaveToolStripButton_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void OpenToolStripButton_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void MoveToolButton_Click(object sender, EventArgs e)
        {
            SetTool(MouseTool.Move);
        }

        private void PlacementToolButton_Click(object sender, EventArgs e)
        {
            SetTool(MouseTool.Placment);
        }

        private void BrickDeleteToolButton_Click(object sender, EventArgs e)
        {
            SetTool(MouseTool.Removal);
        }

        private void SelectToolButton_Click(object sender, EventArgs e)
        {
            SetTool(MouseTool.Select);
        }

        private void ShapeEditorToolStripButton_Click(object sender, EventArgs e)
        {
            shapeEditor.Show();
        }

        private void spriteEditortoolStripButton_Click(object sender, EventArgs e)
        {
            animatedSpriteEditor.Show();
        }

        //
        // Level Tab
        //
        private void LevelBGColorDisplay_DoubleClick(object sender, EventArgs e)
        {
            ColorDialog.ShowDialog();

            Console.AppendText("Level Background Color changed: " + LevelBGColorDisplay.BackColor.ToString()
                + " -> " + ColorDialog.Color.ToString() + "\n");

            LevelBGColorDisplay.BackColor = ColorDialog.Color;

            level.levelAttribs.bgColor = ColorDialog.Color;
            ReDrawLevel();

            DataChanged = true;
        }

        private void FontSelectButton_Click(object sender, EventArgs e)
        {
            fontSelect.SetSelection(FontTextBox.Text);
            fontSelect.FontSize = (int)FontSizeNumericUpDown.Value;

            if (fontSelect.ShowDialog() != DialogResult.OK)
                return;

            FontTextBox.Text = fontSelect.selectionDir;
            FontSizeNumericUpDown.Value = fontSelect.FontSize;

            level.levelAttribs.fontFile = fontSelect.selectionDir;
            level.levelAttribs.fontSize = fontSelect.FontSize;

            DataChanged = true;
        }

        private void LevelNameTextBox_TextChanged(object sender, EventArgs e)
        {
            level.levelAttribs.levelName = LevelNameTextBox.Text;

            DataChanged = true;
        }

        private void FontSizeNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            level.levelAttribs.fontSize = (int)FontSizeNumericUpDown.Value;

            DataChanged = true;
        }

        //
        // Console
        //
        private void ConsoleInputEnterButton_Click(object sender, EventArgs e)
        {
            ConsoleInputEnter();
        }

        private void ConsoleInput_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
                ConsoleInputEnter();
        }

        // 
        // Brickgroup tab
        //
        private void BrickGroupsListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            ignoreDataChange = true;

            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                BrickWidthNumericUpDown.Value = g.BrickWidth;
                BrickHeightNumericUpDown.Value = g.BrickHeight;
                PointsPerHitNumericUpDown.Value = g.Attribs.pointsPerHit;
                PointsPerKillNumericUpDown.Value = g.Attribs.pointsPerKill;
                HitPointsNumericUpDown.Value = g.Attribs.brickHP;
                BrickGroupNameTextBox.Text = g.Attribs.name;

                ExpandImageFileList(g.Attribs.imageFiles, g.Attribs.brickHP);
                ExpandImageList(g.Attribs.brickHP);

                ImageCheckedListBox.ClearSelected();

                UpdateImageFileList(g.Attribs.imageFiles);

                level.SelectCollection(g);

                PointsListBox.DataSource = g.bricks;
                PointsListBox.DisplayMember = "DisplayPoint";
                PointsCountLabel.Text = Convert.ToString(g.bricks.Count);

                grid.CellWidth = g.Attribs.brickWidth;
                grid.CellHeight = g.Attribs.brickHeight;

                CellWidthNumericUpDown.Value = g.Attribs.brickWidth;
                CellHeightNumericUpDown.Value = g.Attribs.brickHeight;
            }
            else
            {
                BrickWidthNumericUpDown.Value = BrickWidthNumericUpDown.Minimum;
                BrickHeightNumericUpDown.Value = BrickHeightNumericUpDown.Minimum;
                PointsPerHitNumericUpDown.Value = PointsPerHitNumericUpDown.Minimum;
                PointsPerKillNumericUpDown.Value = PointsPerKillNumericUpDown.Minimum;
                HitPointsNumericUpDown.Value = HitPointsNumericUpDown.Minimum;
                BrickGroupNameTextBox.Text = "";

                ImageCheckedListBox.Items.Clear();

                level.SelectCollection(null);

                PointsListBox.DataSource = null;
                PointsCountLabel.Text = "0";
            }

            ReDrawLevel();

            ignoreDataChange = false;
        }

        private void BrickWidthNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                g.BrickWidth = (int)BrickWidthNumericUpDown.Value;
                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void BrickHeightNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                g.BrickHeight = (int)BrickHeightNumericUpDown.Value;
                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void PointsPerHitNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                g.Attribs.pointsPerHit = (int)PointsPerHitNumericUpDown.Value;

                DataChanged = true;
            }
        }

        private void PointsPerKillNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                g.Attribs.pointsPerKill = (int)PointsPerKillNumericUpDown.Value;

                DataChanged = true;
            }
        }

        private void HitPointsNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                g.Attribs.brickHP = (int)HitPointsNumericUpDown.Value;
               
                ExpandImageFileList(g.Attribs.imageFiles, g.Attribs.brickHP);
                ExpandImageList(g.Attribs.brickHP);

                UpdateImageFileList(g.Attribs.imageFiles);

                DataChanged = true;
            }
        }

        private void BrickGroupSelectImageButton_Click(object sender, EventArgs e)
        {
            if (ImageNameTextBox.Text != "")
                imageSelect.SetSelection(ImageNameTextBox.Text);

            if (imageSelect.ShowDialog(this) != DialogResult.OK)
                return;

            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                ImageNameTextBox.Text = imageSelect.selectionDir;
                ImageCheckedListBoxItem item = (ImageCheckedListBoxItem)ImageCheckedListBox.SelectedItem;
                
                if (item == null)
                    return;

                item.imageName = imageSelect.selectionName;
                item.imagePath = imageSelect.selectionDir;

                int index = item.index;

                g.SetImage(index, imageSelect.selectionDir);

                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void ImageCheckedListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            ignoreDataChange = true;

            ImageNameTextBox.Text = "";

            int index = ImageCheckedListBox.SelectedIndex;
            if (index >= 0)
            {
                ImageCheckedListBoxItem item = (ImageCheckedListBoxItem)ImageCheckedListBox.SelectedItem;

                if (item.imagePath == null)
                {
                    ImageNameTextBox.Text = "(empty)";
                }
                else
                {
                    ImageNameTextBox.Text = item.imagePath;
                }
            }

            ignoreDataChange = false;
        }

        private void ImageCheckedListBox_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if (ignoreItemCheckedEvent) 
                return;

            ImageCheckedListBoxItem item = (ImageCheckedListBoxItem)ImageCheckedListBox.Items[e.Index];
            if (item.index == 0)
            {
                e.NewValue = CheckState.Checked;
                return;
            }

            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                if (e.NewValue == CheckState.Checked)
                {
                    g.SetImage(e.Index, item.imagePath);
                }
                else
                {
                    String val = "";

                    int last = e.Index - 1;
                    if (last >= 0)
                        val = g.Attribs.imageFiles[last];

                    g.SetImage(e.Index, val);
                }

                // propagates the change up the image list
                int len = ImageCheckedListBox.Items.Count - e.Index;
                for (int i = e.Index + 1; i < len; i++)
                {
                    if (!ImageCheckedListBox.GetItemChecked(i))
                        g.SetImage(i, item.imagePath);
                    else
                        break;
                }

                UpdateImageFileList(g.Attribs.imageFiles);
            }

            DataChanged = true;
        }

        private void AddBrickGroup_Click(object sender, EventArgs e)
        {
            level.CreateNewBrickGroup();

            DataChanged = true;
        }

        private void RemoveBrickGroup_Click(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                level.brickGroups.Remove(g.Attribs);
                level.brickCollections.Remove(g);

                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void BrickGroupNameTextBox_TextChanged(object sender, EventArgs e)
        {

        }

        private void BrickGroupNameTextBox_Leave(object sender, EventArgs e)
        {
            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                g.Name = BrickGroupNameTextBox.Text;
                level.brickCollections.ResetItem(BrickGroupsListBox.SelectedIndex);

                DataChanged = true;
            }
        }

        //
        // Point Distro settings
        //
        private void PointsListBox_MouseClick(object sender, MouseEventArgs e)
        {
            ignoreDataChange = true;

            Level.Brick b = (Level.Brick)PointsListBox.SelectedItem;
            if (b != null)
            {
                level.SelectBrick(b);
                PointXNumericUpDown.Value = b.position.X;
                PointYNumericUpDown.Value = b.position.Y;
                ReDrawLevel();
            }

            ignoreDataChange = false;
        }

        private void PointXNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.Brick b = (Level.Brick)PointsListBox.SelectedItem;
            if (b != null)
            {
                b.MoveToGamePosition(new Point((int)PointXNumericUpDown.Value, b.position.Y));
                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void PointXNumericUpDown_Leave(object sender, EventArgs e)
        {
            Level.Brick b = (Level.Brick)PointsListBox.SelectedItem;
            if (b != null)
            {
                b.owner.bricks.ResetItem(b.attribIndex);
                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void PointYNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            Level.Brick b = (Level.Brick)PointsListBox.SelectedItem;
            if (b != null)
            {
                b.MoveToGamePosition(new Point(b.position.X, (int)PointYNumericUpDown.Value));
                ReDrawLevel();

                DataChanged = true;
            }
        }

        private void PointYNumericUpDown_Leave(object sender, EventArgs e)
        {
            Level.Brick b = (Level.Brick)PointsListBox.SelectedItem;
            if (b != null)
            {
                b.owner.bricks.ResetItem(b.attribIndex);
                ReDrawLevel();
            }
        }

        private void RemovePointButton_Click(object sender, EventArgs e)
        {
            Level.Brick b = (Level.Brick)PointsListBox.SelectedItem;
            if (b != null)
            {
                int nextIndex = b.attribIndex;

                Level.BrickCollection g = b.owner;
                g.RemoveBrick(b);

                if (nextIndex == g.bricks.Count)
                    PointsListBox.SelectedIndex = nextIndex - 1;
                else
                    PointsListBox.SelectedIndex = nextIndex;

                level.SelectBrick((Level.Brick)PointsListBox.SelectedItem);

                ReDrawLevel();

                DataChanged = true;
            }
        }

        // 
        // Paddle tab
        //
        private void PaddleImageSelectButton_Click(object sender, EventArgs e)
        {
            if (imageSelect.ShowDialog() != DialogResult.OK)
                return;

            PaddleImageTextBox.Text = imageSelect.selectionDir;
            level.PaddleImage = imageSelect.selectionDir;

            ReDrawLevel();

            DataChanged = true;
        }

        private void PaddleWidthNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            level.PaddleWidth = (int)PaddleWidthNumericUpDown.Value;
            ReDrawLevel();

            DataChanged = true;
        }

        private void PaddleHeightNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            level.PaddleHeight = (int)PaddleHeightNumericUpDown.Value;
            ReDrawLevel();

            DataChanged = true;
        }

        private void PaddleXNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            level.PaddleX = (int)PaddleXNumericUpDown.Value;
            ReDrawLevel();

            DataChanged = true;
        }

        private void PaddleYNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            level.PaddleY = (int)PaddleYNumericUpDown.Value;
            ReDrawLevel();

            DataChanged = true;
        }

        private void PaddleShapeSelectButton_Click(object sender, EventArgs e)
        {
            if (shapeSelect.ShowDialog() != DialogResult.OK)
                return;

            PaddleShapeTextBox.Text = shapeSelect.selectionDir;
            level.paddleAttribs.shapeFile = shapeSelect.selectionDir;
        }

        private void defaultPaddleAttribs_Click(object sender, EventArgs e)
        {
            PaddleXNumericUpDown.Value = (decimal)0;
            PaddleYNumericUpDown.Value = (decimal)-210;
            PaddleWidthNumericUpDown.Value = (decimal)128;
            PaddleHeightNumericUpDown.Value = (decimal)74;
            PaddleShapeTextBox.Text = "paddle.shape";
            level.paddleAttribs.shapeFile = "paddle.shape";
            PaddleImageTextBox.Text = "board.png";
            level.PaddleImage = "board.png";
        }

        //
        // Ball tab
        //
        private void BallRadiusNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            level.ballAttribs.radius = (int)BallRadiusNumericUpDown.Value;
            ReDrawLevel();

            DataChanged = true;
        }

        private void BallImageSelectButton_Click(object sender, EventArgs e)
        {
            if (imageSelect.ShowDialog() != DialogResult.OK)
                return;

            BallImageTextBox.Text = imageSelect.selectionDir;
            level.BallImage = imageSelect.selectionDir;

            ReDrawLevel();

            DataChanged = true;
        }

        private void defaultBallAtribs_Click(object sender, EventArgs e)
        {
            BallRadiusNumericUpDown.Value = (decimal)9;
            BallImageTextBox.Text = "ball.png";
            level.BallImage = "ball.png";
        }

        //
        // Grid settings
        //
        private void SnapToGridCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            snapToGrid = SnapToGridCheckBox.Checked;
        }

        private void DrawGridCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if (DrawGridCheckBox.Checked)
                level.grid = grid;
            else
                level.grid = null;

            ReDrawLevel();
        }

        private void FixGridCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            fixGrid = FixGridCheckBox.Checked;
        }

        private void RowPaddingNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            int oldValue = grid.CellSpaceY;
            int newValue = (int)RowPaddingNumericUpDown.Value;

            grid.CellSpaceY = newValue;
            if (fixGrid)
            {
                GridOffYNumericUpDown.Value += oldValue - newValue;
            }
            else
            {
                // avoids drawing the level twice if grid fix is on.
                ReDrawLevel();
            }
        }

        private void ColPaddingNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            int oldValue = grid.CellSpaceX;
            int newValue = (int)ColPaddingNumericUpDown.Value;

            grid.CellSpaceX = newValue;
            if (fixGrid)
            {
                GridOffXNumericUpDown.Value += oldValue - newValue;
            }
            else
            {
                // avoids drawing the level twice if grid fix is on.
                ReDrawLevel();
            }
        }

        private void GridOffXNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            grid.OffsetX = (int)GridOffXNumericUpDown.Value;
            ReDrawLevel();
        }

        private void GridOffYNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            grid.OffsetY = (int)GridOffYNumericUpDown.Value;
            ReDrawLevel();
        }

        private void CellWidthNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            grid.CellWidth = (int)CellWidthNumericUpDown.Value;
            ReDrawLevel();
        }

        private void CellHeightNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            grid.CellHeight = (int)CellHeightNumericUpDown.Value;
            ReDrawLevel();
        }

        private void ResetGridButton_Click(object sender, EventArgs e)
        {
            RowPaddingNumericUpDown.Value = 0;
            ColPaddingNumericUpDown.Value = 0;

            GridOffXNumericUpDown.Value = 0;
            GridOffYNumericUpDown.Value = 0;

            Level.BrickCollection g = GetSelectedBrickGroup();
            if (g != null)
            {
                CellWidthNumericUpDown.Value = g.Attribs.brickWidth;
                CellHeightNumericUpDown.Value = g.Attribs.brickHeight;
            }
            else
            {
                CellWidthNumericUpDown.Value = 64;
                CellHeightNumericUpDown.Value = 16;
            }
        }

        //
        // Main window
        //
        private void MainWindow_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.M)
            {
                SetTool(MouseTool.Move);
            }
            else if (e.KeyCode == Keys.S)
            {
                SetTool(MouseTool.Select);
            }
            else if (e.KeyCode == Keys.P)
            {
                SetTool(MouseTool.Placment);
            }
            else if (e.KeyCode == Keys.R)
            {
                SetTool(MouseTool.Removal);
            }
        }

        private void MainWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            if(e.CloseReason == CloseReason.UserClosing || e.CloseReason == CloseReason.ApplicationExitCall)
                e.Cancel = !PromptDataChanged();
        }

        #endregion

        #region Editor Data

        private bool dataChanged = false;
        private bool DataChanged
        {
            set 
            { 
                if(!ignoreDataChange)
                {
                    if(value && !dataChanged)
                    {
                        Text = Text + "*";
                    }
                    else if(!value && dataChanged)
                    {
                        Text = WindowName + currentLevelFile;
                    }

                    dataChanged = true;
                }
            }
        }

        bool ignoreDataChange = false;

        // mouse move stuff
        private Level.Brick dragBrick = null;
        private bool mouseDown = false;

        private bool snapToGrid = false;
        private bool fixGrid = false;
        private Grid grid = new Grid();

        private MouseTool currentTool = MouseTool.Select;

        private static MainWindow window = null;

        private static System.Diagnostics.Process gameProcess = null;
        private delegate void ConsoleWriteCallback(String txt);

        private String currentLevelFile = "Untitled.xml";
        private Level level = new Level();
        private bool levelLoaded = false;

        private ImageSelectionDialog imageSelect = new ImageSelectionDialog();
        private FontSelectionDialog fontSelect = new FontSelectionDialog();
        private ShapeSelectionDialog shapeSelect = new ShapeSelectionDialog();

        private ShapeEditor shapeEditor = new ShapeEditor();
        private AnimatedSpriteEditor animatedSpriteEditor = new AnimatedSpriteEditor();

        private bool ignoreItemCheckedEvent = false;

        public class ImageCheckedListBoxItem
        {
            public override string ToString()
            {
                return listName;
            }

            public int index;

            public String listName = null;

            public String imageName = null;
            public String imagePath = null;
        }

        #endregion
    }
}
