﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.ComponentModel;
using System.IO;

namespace LevelEditor
{
    class Shape
    {
        public class ShapeData
        {
            public String DisplayName
            {
                get
                {
                    switch (type)
                    {
                        case Type.Box:
                            return "Box";
                        case Type.Circle:
                            return "Circle";
                        case Type.Poly:
                            return "Polygon";
                    }

                    return "Unknown Type";
                }
            }

            public enum Type
            {
                Poly = 0,
                Circle,
                Box
            }

            public Type type
            {
                get { return shapeType; }
                set { shapeType = value; recalculateBBox = true; recalculateTransform = true; }
            }

            private Type shapeType = Type.Poly;

            // poly data
            public PointF[] Points
            {
                get { return points; }
                set { points = value; recalculateBBox = true; recalculateTransform = true; }
            }
            public int Count
            {
                get { return count; }
                set { count = value; recalculateBBox = true; recalculateTransform = true; }
            }

            // circle data
            public float Radius
            {
                get { return radius; }
                set { radius = value; recalculateBBox = true; recalculateTransform = true; }
            }

            // box data
            public float W
            {
                get { return w; }
                set { w = value; recalculateBBox = true; recalculateTransform = true; }
            }
            public float H
            {
                get { return h; }
                set { h = value; recalculateBBox = true; recalculateTransform = true; }
            }

            // transformation data
            public float Tx
            {
                get { return tx; }
                set { tx = value; recalculateBBox = true; recalculateTransform = true; }
            }
            public float Ty
            {
                get { return ty; }
                set { ty = value; recalculateBBox = true; recalculateTransform = true; }
            }
            public float Sx
            {
                get { return sx; }
                set { sx = value; recalculateBBox = true; recalculateTransform = true; }
            }
            public float Sy
            {
                get { return sy; }
                set { sy = value; recalculateBBox = true; recalculateTransform = true; }
            }
            public float A
            {
                get { return a; }
                set { a = value; recalculateBBox = true; recalculateTransform = true; }
            }

            // poly data
            private PointF[] points = 
            { 
                new PointF(1.0f, -1.0f),
                new PointF(0.0f, 1.0f),
                new PointF(-1.0f, -1.0f)
            };

            private int count = 3;

            // circle data
            private float radius = 0.0f;

            // box data
            private float w = 0.0f;
            private float h = 0.0f;

            // transformation data
            private float tx = 0.0f;
            private float ty = 0.0f;
            private float sx = 1.0f;
            private float sy = 1.0f;
            private float a = 0.0f;

            public static float DegToRad = (float)Math.PI / 180.0f;
            public static float RadToDeg = 180.0f / (float)Math.PI;

            private bool recalculateTransform = true;
            private bool recalculateBBox = true;

            private Matrix transform = new Matrix();
            private RectangleF boundingBox = new RectangleF();

            public enum WindingOrder
            {
                CW = 0,
                CCW,
                INVALID
            };

            public WindingOrder GetWindingOrder()
            {
                float area = 0.0f;
                for (int i = 0; i < (points.Length - 1); i++)
                {
                    area = area + (points[i].X * points[i + 1].Y) - (points[i + 1].X * points[i].Y);  
                }

                // the actual area is abs(area)/2

                if (area < 0.0f)
                    return WindingOrder.CW;
                else if(area > 0.0f)
                    return WindingOrder.CCW;
                return WindingOrder.INVALID;
            }

            public enum Concavity
            {
                CONCAVE = 0,
                CONVEX,
                INVALID
            };

            public Concavity GetConcavity()
            {
                int n = points.Length;
                if(n < 3)
                    return Concavity.INVALID;

                bool pos = false;
                bool neg = false;

                // computes the cross product between each adjacent edge and makes not of the sign
                for(int i = 0; i < n; i++)
                {
                    int j = (i + 1) % n;
                    int k = (i + 2) % n;
                    float cross = (points[j].X - points[i].X) * (points[k].Y - points[j].Y) -
                                  (points[j].Y - points[i].Y) * (points[k].X - points[j].X);

                    if (cross < 0.0f)
                        pos = true;
                    else if (cross > 0.0f)
                        neg = true;
                    if (pos && neg)
                        return Concavity.CONCAVE;
                }

                if (pos || neg)
                    return Concavity.CONVEX;

                return Concavity.INVALID;
            }

            public void SwitchWindingOrder()
            {
                List<PointF> newPoints = new List<PointF>(points);
                newPoints.Reverse();

                points = newPoints.ToArray();
            }

            private float SignedTriangleArea(PointF a, PointF b, PointF c)
            {
                return (a.X - c.X) * (b.Y - c.Y) * (a.Y - c.Y) * (b.X - c.X);
            }

            // Adapted from Christer Ericson, RealTime Collision Detection pg 152
            public bool LineLineIntersection(PointF a, PointF b, PointF c, PointF d, ref float t)
            {
                // Gets the winding order of the two triangles.  CCW if sign of area is +
                float a1 = SignedTriangleArea(a, b, d);
                float a2 = SignedTriangleArea(a, b, c);

                if (a1 * a2 < 0.0f)
                {
                    float a3 = SignedTriangleArea(c, d, a);
                    float a4 = a3 + a2 - a1;
                    if (a3 * a4 < 0.0f)
                    {
                        t = a3 / (a3 - a4);
                        return true;
                    }
                }

                return false;
            }

            // Not really working :(
            public bool IsSelfIntersecting()
            {
                int n = points.Length;
                if (n < 3)
                    return false;

                for (int i = 0; i < n; i++)
                {
                    int a = i;
                    int b = (i + 1) % n;

                    for (int j = 0; j < n; j++)
                    {
                        if (j != i)
                        {
                            int c = j;
                            int d = (j + 1) % n;

                            float t = 0.0f;
                            if (LineLineIntersection(points[a], points[b], points[c], points[d], ref t))
                            {
                                return true;
                            }
                        }
                    }
                }
                return false;
            }

            public int AddVertex(PointF p)
            {
                PointF[] newPoints = new PointF[points.Length + 1];
                for (int i = 0; i < points.Length; i++)
                {
                    newPoints[i] = points[i];
                }
                newPoints[points.Length] = p;
                points = newPoints;

                return newPoints.Length - 1;
            }

            public void RemovePoint(int index)
            {
                List<PointF> newPoints = new List<PointF>();
                for (int i = 0; i < points.Length; i++)
                {
                    if (i != index)
                        newPoints.Add(points[i]);
                }
                points = newPoints.ToArray();
            }

            public Matrix GetTransform()
            {
                if (!recalculateTransform)
                    return transform;

                Matrix M = new Matrix();
                M.Scale(sx, sy);
                M.Translate(tx, ty);
                M.Rotate(a * RadToDeg);


                transform = M;
                recalculateTransform = false;

                return transform;
            }

            public RectangleF GetBoundingBox()
            {
                if (!recalculateBBox)
                    return boundingBox;

                RectangleF bbox = new RectangleF(0.0f, 0.0f, 0.0f, 0.0f);

                switch (type)
                {
                    case Type.Circle:
                        bbox.X = tx - radius;
                        bbox.Y = ty + radius;
                        bbox.Width = 2 * radius;
                        bbox.Height = 2 * radius;
                        break;

                    case Type.Box:
                        {
                            float w2 = w / 2.0f;
                            float h2 = h / 2.0f;

                            PointF[] boxPoints = new PointF[4];
                            boxPoints[0].X = -w2; boxPoints[0].Y = -h2;
                            boxPoints[1].X = w2; boxPoints[1].Y = -h2;
                            boxPoints[2].X = w2; boxPoints[2].Y = h2;
                            boxPoints[3].X = -w2; boxPoints[3].Y = h2;

                            Matrix M = GetTransform();
                            M.TransformPoints(boxPoints);

                            float minx = float.MaxValue;
                            float miny = float.MaxValue;
                            float maxx = float.MinValue;
                            float maxy = float.MinValue;

                            foreach (PointF p in boxPoints)
                            {
                                minx = Math.Min(p.X, minx);
                                miny = Math.Min(p.Y, miny);

                                maxx = Math.Max(p.X, maxx);
                                maxy = Math.Max(p.Y, maxy);
                            }

                            bbox.X = minx;
                            bbox.Y = maxy;
                            bbox.Width = maxx - minx;
                            bbox.Height = maxy - miny;
                        }
                        break;

                    case Type.Poly:
                        {
                            float minx = float.MaxValue;
                            float miny = float.MaxValue;
                            float maxx = float.MinValue;
                            float maxy = float.MinValue;

                            Matrix M = GetTransform();
                            PointF[] transPoints = (PointF[])points.Clone();
                            M.TransformPoints(transPoints);

                            foreach (PointF p in transPoints)
                            {
                                minx = Math.Min(p.X, minx);
                                miny = Math.Min(p.Y, miny);

                                maxx = Math.Max(p.X, maxx);
                                maxy = Math.Max(p.Y, maxy);
                            }

                            bbox.X = minx;
                            bbox.Y = maxy;
                            bbox.Width = maxx - minx;
                            bbox.Height = maxy - miny;
                        }
                        break;
                }

                boundingBox = bbox;
                recalculateBBox = false;

                return boundingBox;
            }

            public PointF GetCenterPoint()
            {
                RectangleF bbox = GetBoundingBox();
                return new PointF(bbox.X + bbox.Width / 2.0f, bbox.Y - bbox.Height / 2.0f);
            }

            public void DrawTransfomed(Graphics gfx, Pen pen)
            {
                GraphicsState state = gfx.Save();

                gfx.MultiplyTransform(GetTransform());
                Draw(gfx, pen);

                gfx.Restore(state);
            }

            public void Draw(Graphics gfx, Pen pen)
            {
                switch (type)
                {
                    case Type.Circle:
                        gfx.DrawEllipse(pen, -radius, -radius, radius * 2.0f, radius * 2.0f);
                        break;
                    case Type.Box:
                        gfx.DrawRectangle(pen, -w / 2.0f, -h / 2.0f, w, h);
                        break;
                    case Type.Poly:
                        if(points != null)
                            gfx.DrawPolygon(pen, points);
                        break;
                }
            }
        }

        public class InvalidShapeFileException : Exception
        {
            public InvalidShapeFileException(String m) { msg = m; }
            public String msg;
        }

        public BindingList<ShapeData> shapes = new BindingList<ShapeData>();

        public RectangleF GetBoundingBox()
        {
            if (shapes.Count == 0)
                return new RectangleF(0.0f, 0.0f, 0.0f, 0.0f);

            float minx = float.MaxValue;
            float miny = float.MaxValue;
            float maxx = float.MinValue;
            float maxy = float.MinValue;
            foreach (ShapeData shape in shapes)
            {
                RectangleF bb = shape.GetBoundingBox();

                minx = Math.Min(minx, bb.X);
                miny = Math.Min(miny, bb.Y - bb.Height);

                maxx = Math.Max(maxx, bb.X + bb.Width);
                maxy = Math.Max(maxy, bb.Y);
            }

            return new RectangleF(minx, maxy, maxx - minx, maxy - miny);
        }

        public void DrawBoundingBoxes(Graphics gfx)
        {
            foreach (ShapeData s in shapes)
            {
                RectangleF bbox = s.GetBoundingBox();

                gfx.DrawRectangle(outlinePen,
                    bbox.X,
                    bbox.Y - bbox.Height, 
                    bbox.Width, 
                    bbox.Height);
            }
        }

        public void Draw(Graphics gfx)
        {
            foreach (ShapeData s in shapes)
            {
                s.DrawTransfomed(gfx, defaultPen);
            }

            RectangleF bbox = GetBoundingBox();
        }

        private Pen outlinePen = new Pen(Color.HotPink, 0.001f);

        public static Pen defaultPen = new Pen(Color.Green, 0.001f);

        private static char[] deliminators = new char[] { ' ' };

        public static Shape LoadShapeFromFile(String file)
        {
            Shape s = new Shape();

            StreamReader f = File.OpenText(file);
            while (!f.EndOfStream)
            {
                String line = f.ReadLine();
                String[] data = line.Split(deliminators, StringSplitOptions.RemoveEmptyEntries);

                if (data.Length > 0)
                {
                    if (data[0] == "poly")
                    {
                        if(data.Length != 7)
                            throw new InvalidShapeFileException("poly command is incomplete");

                        ShapeData shape = new ShapeData();
                        shape.type = ShapeData.Type.Poly;
                        shape.Count = int.Parse(data[1]);
                        shape.Tx = float.Parse(data[2]);
                        shape.Ty = float.Parse(data[3]);
                        shape.Sx = float.Parse(data[4]);
                        shape.Sy = float.Parse(data[5]);
                        shape.A = float.Parse(data[6]);

                        shape.Points = new PointF[shape.Count];
                        for (int i = 0; i < shape.Count; i++)
                        {
                            if(f.EndOfStream)
                                throw new InvalidShapeFileException("poly command is incomplete");

                            String str = f.ReadLine();
                            String[] p = str.Split(deliminators, StringSplitOptions.RemoveEmptyEntries);

                            if(p.Length != 2)
                                throw new InvalidShapeFileException("poly command is incomplete");

                            PointF point = new PointF();
                            point.X = float.Parse(p[0]);
                            point.Y = float.Parse(p[1]);
                            shape.Points[i] = point;
                        }

                        s.shapes.Add(shape);
                    }
                    else if (data[0] == "circle")
                    {
                        if (data.Length != 4)
                            throw new InvalidShapeFileException("circle command is incomplete");

                        ShapeData shape = new ShapeData();
                        shape.type = ShapeData.Type.Circle;
                        shape.Radius = float.Parse(data[1]);
                        shape.Tx = float.Parse(data[2]);
                        shape.Ty = float.Parse(data[3]);
                        s.shapes.Add(shape);
                    }
                    else if(data[0] == "box")
                    {
                        if (data.Length != 8)
                            throw new InvalidShapeFileException("box command is incomplete");

                        ShapeData shape = new ShapeData();
                        shape.type = ShapeData.Type.Box;
                        shape.W = float.Parse(data[1]);
                        shape.H = float.Parse(data[2]);
                        shape.Tx = float.Parse(data[3]);
                        shape.Ty = float.Parse(data[4]);
                        shape.Sx = float.Parse(data[5]);
                        shape.Sy = float.Parse(data[6]);
                        shape.A = float.Parse(data[7]);
                        s.shapes.Add(shape);
                    }
                }
            }

            f.Close();

            return s;
        }

        public static void SaveShapeToFile(Shape s, String file)
        {
            StreamWriter f = new StreamWriter(File.OpenWrite(file));
            
            foreach (ShapeData shape in s.shapes)
            {
                switch (shape.type)
                {
                    case ShapeData.Type.Circle:
                        f.WriteLine("circle {0} {1} {2}", shape.Radius, shape.Tx, shape.Ty);
                        break;
                    case ShapeData.Type.Box:
                        f.WriteLine("box {0} {1} {2} {3} {4} {5} {6}",
                            shape.W,
                            shape.H,
                            shape.Tx,
                            shape.Ty,
                            shape.Sx,
                            shape.Sy,
                            shape.A);
                        break;
                    case ShapeData.Type.Poly:
                        f.WriteLine("poly {0} {1} {2} {3} {4} {5}",
                            shape.Count,
                            shape.Tx,
                            shape.Ty,
                            shape.Sx,
                            shape.Sy,
                            shape.A);
                        for (int i = 0; i < shape.Count; i++)
                        {
                            f.WriteLine("{0} {1}", shape.Points[i].X, shape.Points[i].Y);
                        }
                        break;
                }
            }

            f.Close();
        }
    }
}
