﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;

namespace LevelEditor
{
    public partial class ShapeEditor : Form
    {
        public static String AppName = "BrickOut |ReDuX| Shape Editor";

        public ShapeEditor()
        {
            InitializeComponent();
        }

        private enum Tool
        {
            Select = 0,
            Vertex,

            ToolCount
        };

        #region Editor methods

        public void LoadShape(String file)
        {
            try
            {
                loadedShape = Shape.LoadShapeFromFile(file);

                ShapeListBox.DataSource = loadedShape.shapes;
            }
            catch (Shape.InvalidShapeFileException)
            {
                loadedShape = null;
            }
        }

        private bool Open()
        {
            if (!PromptChangedData())
                return false;

            if (openFileDialog.ShowDialog() != DialogResult.OK)
                return false;

            currFilePath = openFileDialog.FileName;
            currFileName = Path.GetFileName(currFilePath);

            Text = AppName + " - " + currFileName;

            selectedShape = null;
            selectedShapeIndex = -1;

            LoadShape(currFilePath);

            RedrawShapeViewer();

            fileLoaded = true;

            return true;
        }

        private bool Save()
        {
            if (!fileLoaded)
            {
                saveFileDialog.FileName = currFileName;

                if (saveFileDialog.ShowDialog() != DialogResult.OK)
                    return false;

                currFilePath = saveFileDialog.FileName;
                currFileName = Path.GetFileName(currFilePath);
            }

            Shape.SaveShapeToFile(loadedShape, currFilePath);

            dataChanged = false;

            fileLoaded = true;
            return true;
        }

        private void New()
        {
            loadedShape = new Shape();

            selectedShape = null;
            selectedShapeIndex = -1;
            selectedVertexIndex = -1;

            ShapeListBox.DataSource = null;
            vertexListBox.Items.Clear();

            fileLoaded = false;

            currFileName = "Untitled.shape";
            currFilePath = "";

            Text = AppName + " - " + currFileName;

            RedrawShapeViewer();
        }

        private bool PromptChangedData()
        {
            if (DataChanged)
            {
                DialogResult res = MessageBox.Show("Would you like to save the changes made to " + currFileName + "?",
                    "Save Changes",
                    MessageBoxButtons.YesNoCancel,
                    MessageBoxIcon.Question,
                    MessageBoxDefaultButton.Button3);

                if (res == DialogResult.OK)
                {
                    if (!Save())
                        return false;

                    DataChanged = false;
                }
                else if (res == DialogResult.Cancel)
                {
                    return false;
                }
                else
                {
                    DataChanged = false;
                }
            }

            return true;
        }

        private void RedrawShapeViewer()
        {
            ShapeViewer.Refresh();
        }

        private void SelectShape(int index)
        {
            selectedAxis = Axis.None;
            selectedShapeIndex = index;
            selectedShape = loadedShape.shapes[index];

            selectedVertexIndex = -1;

            SetShapeData(selectedShape);
        }

        private void SetShapeData(Shape.ShapeData shape)
        {
            transX.Value = (decimal)shape.Tx;
            transY.Value = (decimal)shape.Ty;

            if (shape.type == Shape.ShapeData.Type.Circle)
            {
                scaleX.Enabled = false;
                scaleY.Enabled = false;
                angle.Enabled = false;

                circleTab.Enabled = true;
                boxTab.Enabled = false;
                polyTab.Enabled = false;

                radius.Value = (decimal)shape.Radius;

                shapeTypeComboBox.SelectedItem = "Circle";
            }
            else
            {
                scaleX.Enabled = true;
                scaleY.Enabled = true;
                angle.Enabled = true;

                circleTab.Enabled = false;

                scaleX.Value = (decimal)shape.Sx;
                scaleY.Value = (decimal)shape.Sy;
                angle.Value = (decimal)shape.A;

                if (shape.type == Shape.ShapeData.Type.Poly)
                {
                    shapeTypeComboBox.SelectedItem = "Polygon";

                    boxTab.Enabled = false;
                    polyTab.Enabled = true;

                    vertexListBox.Items.Clear();

                    if (shape.Points != null)
                    {
                        for(int i = 0; i < shape.Points.Length; i++)
                        {
                            vertexListBox.Items.Add("Vertex " + i);
                        }
                    }

                    VerifySelectedPoly();
                }
                else
                {
                    shapeTypeComboBox.SelectedItem = "Box";

                    boxTab.Enabled = true;
                    polyTab.Enabled = false;

                    boxWidth.Value = (decimal)shape.W;
                    boxHeight.Value = (decimal)shape.H;
                }
            }
        }

        private void SetTool(Tool t)
        {
            if (t == Tool.Vertex)
            {
                ShapeViewer.Cursor = Cursors.Cross;
            }
            else
            {
                ShapeViewer.Cursor = Cursors.Arrow;
            }

            activeTool = t;
        }

        // Gets the bounding box that holds the active translation axis
        private RectangleF GetAxisBBox()
        {
            RectangleF bbox = new RectangleF();

            if (selectedShape != null)
            {
                float w, h, scaleX, scaleY;
                GetShapeViewerDimensions(out w, out h, out scaleX, out scaleY);

                float axisLength = axisArrowLength * w;

                PointF pos = new PointF();
                if (selectedVertexIndex != -1)
                {
                    pos.X = selectedShape.Points[selectedVertexIndex].X;
                    pos.Y = selectedShape.Points[selectedVertexIndex].Y;
                    Transform(selectedShape.GetTransform(), ref pos);
                }
                else
                {
                    pos = selectedShape.GetCenterPoint();
                }

                bbox.X = pos.X;
                bbox.Y = pos.Y + axisLength;
                bbox.Width = axisLength;
                bbox.Height = axisLength;
            }

            return bbox;
        }

        private bool Inside(PointF p, RectangleF r)
        {
            if(p.X >= r.X && 
               p.X <= (r.X + r.Width) && 
               p.Y <= r.Y &&
               p.Y >= (r.Y - r.Height))
            {
                return true;
            }

            return false;
        }

        private int PickVertex(Shape.ShapeData data, PointF pos)
        {
            float w, h, sx, sy;
            GetShapeViewerDimensions(out w, out h, out sx, out sy);

            float size = vertexSize * w;
            float size2 = size / 2.0f;

            Matrix m = data.GetTransform();
            for (int i = 0; i < data.Points.Length; i++)
            {
                PointF p = Transform(m, data.Points[i]);
                RectangleF r = new RectangleF(p.X - size2, p.Y + size2, size, size);
                if (Inside(pos, r))
                {
                    return i;
                }
            }

            return -1;
        }

        private void VerifySelectedPoly()
        {
            if (selectedShape != null && selectedShape.type == Shape.ShapeData.Type.Poly)
            {
                Shape.ShapeData.Concavity con = selectedShape.GetConcavity();
                switch (con)
                {
                    case Shape.ShapeData.Concavity.CONVEX:
                        concavity.Text = "Convex";
                        concavity.ForeColor = Color.Black;
                        break;
                    case Shape.ShapeData.Concavity.CONCAVE:
                        concavity.Text = "Concave";
                        concavity.ForeColor = Color.Red;
                        break;
                    case Shape.ShapeData.Concavity.INVALID:
                        concavity.Text = "INVALID";
                        concavity.ForeColor = Color.Red;
                        break;
                }

                Shape.ShapeData.WindingOrder win = selectedShape.GetWindingOrder();
                switch (win)
                {
                    case Shape.ShapeData.WindingOrder.CCW:
                        windingOrder.Text = "CCW";
                        windingOrder.ForeColor = Color.Black;
                        break;
                    case Shape.ShapeData.WindingOrder.CW:
                        windingOrder.Text = "CW";
                        windingOrder.ForeColor = Color.Red;
                        break;
                    case Shape.ShapeData.WindingOrder.INVALID:
                        windingOrder.Text = "INVALID";
                        windingOrder.ForeColor = Color.Red;
                        break;
                }
            }
        }

        // gets the half width and height for the ShapeViewer as well as the scale to fit to the shapeViewer
        private void GetShapeViewerDimensions(out float w, out float h, out float scaleX, out float scaleY)
        {
            float zoom = GetZoom();

            // linearly interpolate
            w = minx + (maxx - minx) * zoom;
            h = miny + (maxy - miny) * zoom;

            scaleX = (0.5f * ShapeViewer.Width) / w;
            scaleY = (0.5f * ShapeViewer.Height) / h;
        }

        private float GetZoom()
        {
            return zoomTrackBar.Value / (float)zoomTrackBar.Maximum;
        }

        // Computes the transformation matrix that moves from the forms coordinate system to the shapeviewers
        private Matrix GetFormToViewTransform()
        {
            Matrix trans = new Matrix();

            float w, h, scaleX, scaleY;
            GetShapeViewerDimensions(out w, out h, out scaleX, out scaleY);

            trans.Scale(scaleX, -scaleY);
            trans.Translate(w + centerX, -h + centerY);

            return trans;
        }

        // Computes the transformation matrix without any scaling
        private Matrix GetFormToViewTransformNoScale()
        {
            Matrix trans = new Matrix();

            float w, h, scaleX, scaleY;
            GetShapeViewerDimensions(out w, out h, out scaleX, out scaleY);

            trans.Translate((w + centerX) * scaleX,
                            (-h + centerY) * -scaleY);

            return trans;
        }

        private PointF TransformMouseLocation(Point mouse)
        {
            float w, h, scaleX, scaleY;
            GetShapeViewerDimensions(out w, out h, out scaleX, out scaleY);

            PointF p = new PointF();

            float dx = mouse.X / (float)ShapeViewer.Width;
            float dy = mouse.Y / (float)ShapeViewer.Height;

            float minx = -w - centerX;
            float maxx =  w - centerX;
            float miny =  h - centerY;
            float maxy = -h - centerY;

            p.X = minx + (maxx - minx) * dx;
            p.Y = miny + (maxy - miny) * dy;

            return p;
        }

        private void Transform(Matrix m, ref PointF p)
        {
            float x = m.Elements[0] * p.X + m.Elements[1] * p.Y + m.Elements[4];
            float y = m.Elements[2] * p.X + m.Elements[3] * p.Y + m.Elements[5];

            p.X = x;
            p.Y = y;
        }

        private PointF Transform(Matrix m, PointF p)
        {
            float x = m.Elements[0] * p.X + m.Elements[1] * p.Y + m.Elements[4];
            float y = m.Elements[2] * p.X + m.Elements[3] * p.Y + m.Elements[5];

            return new PointF(x, y);
        }

        private void DrawPoint(Graphics gfx, PointF p, Brush b, Matrix m, float size)
        {
            PointF[] tp = { p };

            m.TransformPoints(tp);

            float s2 = size / 2.0f;
            gfx.FillRectangle(b, tp[0].X - s2, tp[0].Y - s2, size, size);
        }

        private void DrawPoints(Graphics gfx, PointF[] points, Brush b, Matrix m, float size)
        {
            foreach (PointF p in points)
            {
                DrawPoint(gfx, p, b, m, size);
            }
        }

        private void DrawArrow(Graphics gfx, PointF p1, PointF p2, Pen pen)
        {
            gfx.DrawLine(pen, p1, p2);
        }

        #endregion

        #region Events

        //
        // Window load/close
        // 
        private void ShapeEditor_Load(object sender, EventArgs e)
        {
            Text = AppName + " - " + currFileName;

            openFileDialog.InitialDirectory = DataPaths.Paths["Shapes"];
            saveFileDialog.InitialDirectory = DataPaths.Paths["Shapes"];

            ShapeListBox.DataSource = loadedShape.shapes;

            ShapeListBox.SelectedIndex = -1;

            circleTab.Enabled = false;
            boxTab.Enabled = false;
            polyTab.Enabled = false;

            xAxisPen.EndCap = LineCap.Custom;
            yAxisPen.EndCap = LineCap.Custom;
            xAxisPen.CustomEndCap = new AdjustableArrowCap(5.0f, 5.0f);
            yAxisPen.CustomEndCap = new AdjustableArrowCap(5.0f, 5.0f);
        }

        private void ShapeEditor_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.ApplicationExitCall || e.CloseReason == CloseReason.UserClosing)
            {
                e.Cancel = true;
                Hide();
            }
        }

        //
        // Paint Events
        // 
        private void ShapeViewer_Paint(object sender, PaintEventArgs e)
        {
            // Sets up screen tranform
            float w, h, scaleX, scaleY;
            GetShapeViewerDimensions(out w, out h, out scaleX, out scaleY);

            Matrix trans = GetFormToViewTransform();

            Graphics gfx = ShapeViewer.CreateGraphics();
            gfx.MultiplyTransform(trans);

            if (drawImageCheckBox.Checked && previewImage != null)
            {
                // NOTE: negitive height accounts for the y-axis being flipped
                gfx.DrawImage(previewImage, 
                              previewImageX - previewImageW / 2.0f, 
                              previewImageY + previewImageH / 2.0f, 
                              previewImageW, 
                              -previewImageH);

            }

            if (drawMinorGridLinesCheckBox.Checked)
            {
            }

            if (drawMajorGirdLinesCheckBox.Checked)
            {
                // x-axis
                gfx.DrawLine(majorGridPen, -MAX_X, 0.0f, MAX_X, 0.0f);
                // y-axis
                gfx.DrawLine(majorGridPen, 0.0f, -MAX_Y, 0.0f, MAX_Y);
            }

            foreach (Shape.ShapeData s in loadedShape.shapes)
            {
                if (s != selectedShape)
                    s.DrawTransfomed(gfx, shapePen);
            }

            if (drawBBoxCheckBox.Checked)
            {
                loadedShape.DrawBoundingBoxes(gfx);
            }

            if (selectedShape != null)
            {
                Matrix m = selectedShape.GetTransform();

                selectedShape.DrawTransfomed(gfx, selectedPen);

                if (selectedShape.type == Shape.ShapeData.Type.Poly)
                {
                    GraphicsState state = gfx.Save();

                    float size = vertexSize * w;

                    DrawPoints(gfx, selectedShape.Points, vertexBrush, m, size);

                    if (selectedVertexIndex != -1)
                    {
                        DrawPoint(gfx, selectedShape.Points[selectedVertexIndex], selectedVertexBrush, m, size);
                    }

                    gfx.Restore(state);
                }

                // draws the transform arrows
                PointF axisPos = new PointF();
                if (selectedVertexIndex != -1)
                {
                    axisPos.X = selectedShape.Points[selectedVertexIndex].X;
                    axisPos.Y = selectedShape.Points[selectedVertexIndex].Y;

                    Transform(m, ref axisPos);
                }
                else
                {
                    RectangleF bbox = selectedShape.GetBoundingBox();
                    axisPos.X = bbox.X + bbox.Width / 2.0f;
                    axisPos.Y = bbox.Y - bbox.Height / 2.0f;
                }

                float arrowLen = axisArrowLength * w;
                float boxDim = arrowLen / 4.0f;

                RectangleF axisBox = new RectangleF(axisPos.X, axisPos.Y, boxDim, boxDim);
                gfx.FillRectangle(axisBrush, axisBox);
                gfx.DrawRectangle(outlinePen, axisBox.X, axisBox.Y, axisBox.Width, axisBox.Height);

                // x-axis
                DrawArrow(gfx, axisPos, new PointF(axisPos.X + arrowLen, axisPos.Y), xAxisPen);

                // y-axis
                DrawArrow(gfx, axisPos, new PointF(axisPos.X, axisPos.Y + arrowLen), yAxisPen);

                float circleDim = boxDim / 2.0f;
                RectangleF circle = new RectangleF(axisPos.X - circleDim / 2.0f,
                                                   axisPos.Y - circleDim / 2.0f,
                                                   circleDim,
                                                   circleDim);

                gfx.FillEllipse(Brushes.Green, circle);
            }

            if (drawGridPoints.Checked)
            {
                PointF[] positions =
                {
                    new PointF(        0.0f,         0.0f),
                    new PointF(        0.0f,  h - centerY),
                    new PointF(-w - centerX,         0.0f),
                    new PointF( w - centerX,         0.0f),
                    new PointF(        0.0f, -w - centerY)
                };

                String p1 = "(" + positions[1].X + "," + positions[1].Y + ")";
                String p2 = "(" + positions[2].X + "," + positions[2].Y + ")";
                String p3 = "(" + positions[3].X + "," + positions[3].Y + ")";
                String p4 = "(" + positions[4].X + "," + positions[4].Y + ")";

                // Transforms the points to fit correctly on the screen
                Matrix scale = new Matrix();
                scale.Scale(scaleX, -scaleY);
                scale.TransformPoints(positions);

                GraphicsState state = gfx.Save();
                gfx.ResetTransform();

                // removes scaling
                gfx.MultiplyTransform(GetFormToViewTransformNoScale());

                // corrects the positioning so we don't draw of the screen
                float strH = markerFont.GetHeight();
                SizeF size = gfx.MeasureString(p3, markerFont);

                positions[3].X -= size.Width;
                positions[4].Y -= strH;

                gfx.DrawString("(0,0)", markerFont, fontBrush, positions[0]);
                gfx.DrawString(p1, markerFont, fontBrush, positions[1]);
                gfx.DrawString(p2, markerFont, fontBrush, positions[2]);
                gfx.DrawString(p3, markerFont, fontBrush, positions[3]);
                gfx.DrawString(p4, markerFont, fontBrush, positions[4]);

                gfx.Restore(state);
            }
        }

        //
        // Shape viewer
        //
        private void ShapeViewer_MouseDown(object sender, MouseEventArgs e)
        {
            mouseDown = true;
            mouseButton = e.Button;

            lastPos.X = e.X;
            lastPos.Y = e.Y;
            currPos.X = e.X;
            currPos.Y = e.Y;

            PointF pos = TransformMouseLocation(e.Location);

            if (activeTool == Tool.Vertex)
            {
                if (selectedShape != null && selectedShape.type == Shape.ShapeData.Type.Poly)
                {
                    if (mouseButton == MouseButtons.Left && selectedShape.Points.Length < MaxPolyVertices)
                    {
                        int index = selectedShape.AddVertex(pos);
                        vertexListBox.Items.Add("Vertex " + (selectedShape.Points.Length - 1));
                        selectedShapeIndex = index;

                        VerifySelectedPoly();

                        RedrawShapeViewer();
                    }
                    else if (mouseButton == MouseButtons.Right &&
                             selectedShape.Points.Length > 3)
                    {
                        int index = PickVertex(selectedShape, pos);
                        if (index != -1)
                        {
                            selectedShape.RemovePoint(index);
                            vertexListBox.Items.RemoveAt(vertexListBox.Items.Count - 1);
                            selectedVertexIndex = -1;

                            VerifySelectedPoly();

                            RedrawShapeViewer();
                        }
                    }
                }
            }
            else
            {
                if (mouseButton == MouseButtons.Left)
                {
                    if (selectedShape != null)
                    {
                        RectangleF axisBBox = GetAxisBBox();

                        float axisSize = axisBBox.Width / 4.0f;

                        PointF center = new PointF(axisBBox.X, axisBBox.Y - axisBBox.Height);
                        RectangleF xAxis = new RectangleF(center.X + axisSize,
                                                          center.Y + axisSize / 2.0f,
                                                          axisBBox.Width - axisSize,
                                                          axisSize);

                        RectangleF yAxis = new RectangleF(center.X - axisSize / 2.0f,
                                                          center.Y + axisBBox.Width,
                                                          axisSize,
                                                          axisBBox.Height - axisSize);

                        RectangleF box = new RectangleF(center.X - axisSize / 2.0f,
                                                        center.Y + axisSize,
                                                        1.5f * axisSize,
                                                        1.5f * axisSize);
                        if (Inside(pos, xAxis))
                        {
                            xAxisPen.Color = Color.Red;
                            selectedAxis = Axis.XAxis;
                        }
                        else if (Inside(pos, yAxis))
                        {
                            yAxisPen.Color = Color.Red;
                            selectedAxis = Axis.YAxis;
                        }
                        else if (Inside(pos, box))
                        {
                            xAxisPen.Color = Color.Red;
                            yAxisPen.Color = Color.Red;
                            selectedAxis = Axis.Both;
                        }
                        else if (selectedShape.type == Shape.ShapeData.Type.Poly)
                        {
                            int index = PickVertex(selectedShape, pos);
                            if (index != -1)
                            {
                                xAxisPen.Color = Color.Red;
                                yAxisPen.Color = Color.Red;
                                selectedAxis = Axis.Both;
                                vertexListBox.SelectedIndex = index;
                            }
                        }
                    }
                }
            }
        }

        private void ShapeViewer_MouseEnter(object sender, EventArgs e)
        {
            ShapeViewer.Focus(); // HACK to get C# to send mouse wheel events
        }

        private void ShapeViewer_MouseWheel(object sender, MouseEventArgs e)
        {
            if (e.Delta != 0)
            {
                int inc = -mouseWheelZoomIncrement * (e.Delta / 120); // One wheel notch == 120
                zoomTrackBar.Value = Math.Min(Math.Max(zoomTrackBar.Value + inc, 0),
                                              zoomTrackBar.Maximum);

                RedrawShapeViewer();
            }
        }

        private void ShapeViewer_MouseMove(object sender, MouseEventArgs e)
        {
            if (mouseDown)
            {
                lastPos.X = currPos.X;
                lastPos.Y = currPos.Y;

                currPos.X = e.X;
                currPos.Y = e.Y;

                float w, h, scaleX, scaleY;
                GetShapeViewerDimensions(out w, out h, out scaleX, out scaleY);

                float dx = (currPos.X - lastPos.X) / (float)ShapeViewer.Width;
                float dy = (lastPos.Y - currPos.Y) / (float)ShapeViewer.Height;

                if (mouseButton == MouseButtons.Middle)
                {
                    centerX += gridTranslateSpeed * dx * w;
                    centerY += gridTranslateSpeed * dy * h;

                    RedrawShapeViewer();
                }
                else if (mouseButton == MouseButtons.Left)
                {
                    PointF pos = TransformMouseLocation(e.Location);

                    if (selectedAxis == Axis.Both)
                    {
                        if (selectedVertexIndex != -1)
                        {
                            vertX.Value += (decimal)(dx * 2.0f * w);
                            vertY.Value += (decimal)(dy * 2.0f * h);
                        }
                        else
                        {
                            transX.Value += (decimal)(dx * 2.0f * w);
                            transY.Value += (decimal)(dy * 2.0f * h);
                        }
                    }
                    else if (selectedAxis == Axis.XAxis)
                    {
                        if (selectedVertexIndex != -1)
                            vertX.Value += (decimal)(dx * 2.0f * w);
                        else
                            transX.Value += (decimal)(dx * 2.0f * w);
                    }
                    else if (selectedAxis == Axis.YAxis)
                    {
                        if (selectedVertexIndex != -1)
                            vertY.Value += (decimal)(dy * 2.0f * h);
                        else
                            transY.Value += (decimal)(dy * 2.0f * h);
                    }

                    if (selectedVertexIndex != -1)
                        VerifySelectedPoly();
                }
            }
        }

        private void ShapeViewer_MouseUp(object sender, MouseEventArgs e)
        {
            mouseDown = false;
            selectedAxis = Axis.None;

            xAxisPen.Color = xAxisColor;
            yAxisPen.Color = yAxisColor;

            RedrawShapeViewer();
        }

        //
        // Menu bar
        //
        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog.FileName = currFileName;

            if (saveFileDialog.ShowDialog() != DialogResult.OK)
                return;

            currFilePath = saveFileDialog.FileName;
            currFileName = Path.GetFileName(currFilePath);

            Shape.SaveShapeToFile(loadedShape, currFilePath);

            dataChanged = false;

            fileLoaded = true;
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!PromptChangedData())
                return;

            currFileName = "Untitled.shape";
            currFilePath = "";

            Text = AppName + " - " + currFileName;

            loadedShape = new Shape();
            ShapeListBox.DataSource = loadedShape.shapes;

            selectedShape = null;
            selectedShapeIndex = -1;
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            New();
        }

        //
        // Tool Strip
        //
        private void OpenToolStripButton_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void SaveToolStripButton_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void VertexToolStripButton_Click(object sender, EventArgs e)
        {
            SetTool(Tool.Vertex);
        }

        private void SelectToolStripButton_Click(object sender, EventArgs e)
        {
            SetTool(Tool.Select);
        }

        //
        // Right bar
        //
        private void shapeTypeComboBox_SelectionChangeCommitted(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                String item = (String)shapeTypeComboBox.SelectedItem;
                if (item == "Polygon")
                    selectedShape.type = Shape.ShapeData.Type.Poly;
                else if (item == "Box")
                    selectedShape.type = Shape.ShapeData.Type.Box;
                else
                    selectedShape.type = Shape.ShapeData.Type.Circle;

                SelectShape(loadedShape.shapes.IndexOf(selectedShape));

                loadedShape.shapes.ResetItem(selectedShapeIndex);
                RedrawShapeViewer();
            }
        }

        private void ShapeListBox_Click(object sender, EventArgs e)
        {
            if (ShapeListBox.SelectedIndex != -1)
                SelectShape(ShapeListBox.SelectedIndex);

            RedrawShapeViewer();
        }

        private void AddButton_Click(object sender, EventArgs e)
        {
            Shape.ShapeData data = new Shape.ShapeData();
            loadedShape.shapes.Add(data);

            SelectShape(loadedShape.shapes.Count() - 1);

            RedrawShapeViewer();
        }

        private void RemoveButton_Click(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                int index = selectedShapeIndex;
                loadedShape.shapes.Remove(selectedShape);

                selectedShape = null;
                selectedShapeIndex = -1;

                ShapeListBox.SelectedIndex = -1;

                RedrawShapeViewer();
            }
        }

        //
        // Transform controls
        // 
        private void transX_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.Tx = (float)transX.Value;
                RedrawShapeViewer();
            }
        }

        private void transY_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.Ty = (float)transY.Value;
                RedrawShapeViewer();
            }
        }

        private void scaleX_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.Sx = (float)scaleX.Value;
                RedrawShapeViewer();
            }
        }

        private void scaleY_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.Sy = (float)scaleY.Value;
                RedrawShapeViewer();
            }
        }

        private void angle_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.A = (float)angle.Value;
                RedrawShapeViewer();
            }
        }

        //
        // Polygon Tab
        //
        private void vertexListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedVertexIndex = vertexListBox.SelectedIndex;

                vertX.Value = (decimal)selectedShape.Points[selectedVertexIndex].X;
                vertY.Value = (decimal)selectedShape.Points[selectedVertexIndex].Y;

                RedrawShapeViewer();
            }
        }

        private void vertX_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null && selectedVertexIndex != -1)
            {
                selectedShape.Points[selectedVertexIndex].X = (float)vertX.Value;
                RedrawShapeViewer();
            }
        }

        private void vertY_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null && selectedVertexIndex != -1)
            {
                selectedShape.Points[selectedVertexIndex].Y = (float)vertY.Value;
                RedrawShapeViewer();
            }
        }

        private void switchWindingOrderButton_Click(object sender, EventArgs e)
        {
            if (selectedShape != null && selectedShape.type == Shape.ShapeData.Type.Poly)
            {
                selectedShape.SwitchWindingOrder();
                VerifySelectedPoly();

                RedrawShapeViewer();
            }
        }

        //
        // Circle Tab
        // 
        private void radius_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.Radius = (float)radius.Value;
                RedrawShapeViewer();
            }
        }

        //
        // Box Tab
        //
        private void boxWidth_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.W = (float)boxWidth.Value;
                RedrawShapeViewer();
            }
        }

        private void boxHeight_ValueChanged(object sender, EventArgs e)
        {
            if (selectedShape != null)
            {
                selectedShape.H = (float)boxHeight.Value;
                RedrawShapeViewer();
            }
        }

        //
        // Grid controls
        //
        private void drawMajorGirdLinesCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            RedrawShapeViewer();
        }

        private void drawGridPoints_CheckedChanged(object sender, EventArgs e)
        {
            RedrawShapeViewer();
        }

        private void drawMinorGridLinesCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            RedrawShapeViewer();
        }

        private void drawBBoxCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            RedrawShapeViewer();
        }

        private void snapToGridCheckBox_CheckedChanged(object sender, EventArgs e)
        {

        }

        private void centerGridButton_Click(object sender, EventArgs e)
        {
            centerX = 0.0f;
            centerY = 0.0f;
            RedrawShapeViewer();
        }

        private void zoomTrackBar_Scroll(object sender, EventArgs e)
        {
            RedrawShapeViewer();
        }

        //
        // Preview Image controls
        //
        private void previewImageSelect_Click(object sender, EventArgs e)
        {
            if (imageOpenFileDialog.ShowDialog() != DialogResult.OK)
                return;

            try
            {
                previewImage = Image.FromFile(imageOpenFileDialog.FileName);
                previewImageTextBox.Text = Path.GetFileName(imageOpenFileDialog.FileName);

                if (PreserveAspect && previewImage != null)
                {
                    float aspect = (float)previewImage.Height / previewImage.Width;
                    imageHeight.Value = (decimal)(previewImageW * aspect);
                }

                RedrawShapeViewer();
            }
            catch (OutOfMemoryException)
            {
                previewImage = null;
                previewImageTextBox.Text = "";
            }
            catch (FileNotFoundException)
            {
                previewImage = null;
                previewImageTextBox.Text = "";
            }
        }

        private void drawImageCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if (previewImage != null)
            {
                RedrawShapeViewer();
            }
        }

        private void imageWidth_ValueChanged(object sender, EventArgs e)
        {
            previewImageW = (float)imageWidth.Value;
            if (!ignoreImageChange && PreserveAspect && previewImage != null)
            {
                ignoreImageChange = true;
                float aspect = (float)previewImage.Height / previewImage.Width;
                imageHeight.Value = (decimal)(previewImageW * aspect);
                ignoreImageChange = false;
            }

            if (!ignoreImageChange && drawImageCheckBox.Checked)
                RedrawShapeViewer();
        }

        private void imageHeight_ValueChanged(object sender, EventArgs e)
        {
            previewImageH = (float)imageHeight.Value;
            if (!ignoreImageChange && PreserveAspect && previewImage != null)
            {
                ignoreImageChange = true;
                float aspect = (float)previewImage.Width / previewImage.Height;
                imageWidth.Value = (decimal)(previewImageH * aspect);
                ignoreImageChange = false;
            }

            if (!ignoreImageChange && drawImageCheckBox.Checked)
                RedrawShapeViewer();
        }

        private void imageX_ValueChanged(object sender, EventArgs e)
        {
            previewImageX = (float)imageX.Value;

            if (previewImage != null && drawImageCheckBox.Checked)
            {
                RedrawShapeViewer();
            }
        }

        private void imageY_ValueChanged(object sender, EventArgs e)
        {
            previewImageY = (float)imageY.Value;

            if (previewImage != null && drawImageCheckBox.Checked)
            {
                RedrawShapeViewer();
            }
        }

        private void preserveAspect_CheckedChanged(object sender, EventArgs e)
        {
            if (PreserveAspect && previewImage != null)
            {
                float aspect = (float)previewImage.Height / previewImage.Width;
                imageHeight.Value = (decimal)(previewImageW * aspect);

                if (drawImageCheckBox.Checked)
                    RedrawShapeViewer();
            }
        }

        #endregion

        #region Data

        //
        // Constants
        //
        private const float gridTranslateSpeed = 2.0f;
        private const int mouseWheelZoomIncrement = 10;
        private const float axisArrowLength = 0.25f;
        private const float vertexSize = 0.025f;

        // hard limit enforced by Box2D
        public const int MaxPolyVertices = 8;

        private bool mouseDown = false;
        private MouseButtons mouseButton = MouseButtons.None; // only useful when mouseDown == true
        private Point lastPos = new Point(0, 0);
        private Point currPos = new Point(0, 0);

        private Tool activeTool = Tool.Select;

        private enum Axis
        {
            XAxis = 0,
            YAxis,
            Both,
            None
        };
        private Axis selectedAxis = Axis.None;

        private Shape loadedShape = new Shape();
        private Shape.ShapeData selectedShape = null;
        private int selectedShapeIndex = -1;
        private int selectedVertexIndex = -1;

        private String currFilePath = "";
        private String currFileName = "Untitled.shape";
        private bool fileLoaded = false;

        private Pen shapePen = new Pen(Color.Blue, 0.001f);
        private Pen selectedPen = new Pen(Color.Green, 0.001f);

        private Color xAxisColor = Color.Blue;
        private Color yAxisColor = Color.Yellow;

        private Pen xAxisPen = new Pen(Color.Blue, 0.005f);
        private Pen yAxisPen = new Pen(Color.Yellow, 0.005f);
        private Brush axisBrush = new SolidBrush(Color.FromArgb(128, Color.Gray));
        private Pen outlinePen = new Pen(Color.White, 0.001f);

        private Brush vertexBrush = Brushes.Orange;
        private Brush selectedVertexBrush = Brushes.CornflowerBlue;

        private Pen majorGridPen = new Pen(Color.White, 0.001f);

        private Brush fontBrush = Brushes.White;
        private Font markerFont = new Font("Times", 7.0f);

        private bool DataChanged
        {
            get { return dataChanged; }
            set 
            {
                if (value && !dataChanged)
                    Text += "*";
                else if (!value && dataChanged)
                    Text = AppName + " - " + currFileName;
                   
                DataChanged = value; 
            }
        }
        private bool dataChanged = false;

        private float centerX = 0.0f;
        private float centerY = 0.0f;
        private float maxx = 20.0f;
        private float minx = 0.2f;
        private float maxy = 20.0f;
        private float miny = 0.2f;

        private static float MAX_X = 1000.0f;
        private static float MAX_Y = 1000.0f;

        // Preview image data
        private Image previewImage = null;
        private float previewImageW = 1.0f;
        private float previewImageH = 1.0f;
        private float previewImageX = 0.0f;
        private float previewImageY = 0.0f;
        private bool PreserveAspect
        {
            get { return preserveAspect.Checked; }
        }

        // used so we don't get in an ifinite loop when preserving aspect ratios.
        private bool ignoreImageChange = false; 

        #endregion
    }
}
